"""Utilities for reconstructing protein structures."""

from typing import Iterable, List, Tuple, Union

import numpy as np
import torch

from ..resi_atoms import ATOM_NUMBERING, BB_BUILD_INFO, SC_BUILD_INFO
from .angles import torsion_to_rad
from .pnerf import sn_nerf
from .types import AtomTensor, CoordTensor, TorsionTensor


def place_fourth_atom(
    a_coord: torch.Tensor,
    b_coord: torch.Tensor,
    c_coord: torch.Tensor,
    length: torch.Tensor,
    planar: torch.Tensor,
    dihedral: torch.Tensor,
) -> torch.Tensor:
    """
    Given 3 coords + a length + a planar angle + a dihedral angle, compute a
    fourth coord

    Adapted from the IgFold Implementation (which is under the John's
    Hopkins License):
    https://github.com/Graylab/IgFold/blob/main/igfold/utils/coordinates.py

    :param a_coord: First coordinate
    :type a_coord: torch.Tensor
    :param b_coord: Second coordinate
    :type b_coord: torch.Tensor
    :param c_coord: Third coordinate
    :type c_coord: torch.Tensor
    :param length: Length of the bond
    :type length: torch.Tensor
    :param planar: Planar angle
    :type planar: torch.Tensor
    :param dihedral: Dihedral angle
    :type dihedral: torch.Tensor
    :return: Fourth coordinate
    :rtype: torch.Tensor
    """
    bc_vec = b_coord - c_coord
    bc_vec = bc_vec / bc_vec.norm(dim=-1, keepdim=True)

    n_vec = (b_coord - a_coord).expand(bc_vec.shape).cross(bc_vec)
    n_vec = n_vec / n_vec.norm(dim=-1, keepdim=True)

    m_vec = [bc_vec, n_vec.cross(bc_vec), n_vec]
    d_vec = [
        length * torch.cos(planar),
        length * torch.sin(planar) * torch.cos(dihedral),
        -length * torch.sin(planar) * torch.sin(dihedral),
    ]

    return c_coord + sum(m * d for m, d in zip(m_vec, d_vec))


def get_ideal_backbone_coords(
    n: int = 1, ca_center: bool = True, device: torch.device = "cpu"
) -> torch.Tensor:
    """
    Get idealized backbone ``(N, CA, C, CB)`` coordinates.

    Adapted from the IgFold Implementation (which is under the John's
    Hopkins License):
    https://github.com/Graylab/IgFold/blob/main/igfold/utils/coordinates.py

    :param center: if ``True`` the reference residue is centered on CA,
        otherwise, the center of mass is used. Defaults to ``True``.
    :type CA_centered: bool, optional
    :return: Idealized backbone coordinates
    :rtype: torch.Tensor
    """
    N = torch.tensor([[0, 0, -1.458]], dtype=torch.float, device=device)
    A = torch.tensor([[0, 0, 0]], dtype=torch.float, device=device)
    B = torch.tensor([[0, 1.426, 0.531]], dtype=torch.float, device=device)
    C = place_fourth_atom(
        B,
        A,
        N,
        torch.tensor(2.460),
        torch.tensor(0.615),
        torch.tensor(-2.143),
    )

    coords = torch.cat([N, A, C, B]).float()

    if not ca_center:
        coords -= coords.mean(
            dim=0,
            keepdim=True,
        )

    return coords.unsqueeze(0).repeat(n, 1, 1)


def place_o_coords(coords: AtomTensor) -> AtomTensor:
    """
    Adapted from the IgFold Implementation (which is under the John's
    Hopkins License):
    https://github.com/Graylab/IgFold/blob/main/igfold/utils/coordinates.py

    :param coords: _description_
    :type coords: _type_
    :return: _description_
    :rtype: _type_
    """
    N = coords[:, 0, :]
    A = coords[:, 1, :]
    C = coords[:, 2, :]

    o_coords = place_fourth_atom(
        torch.roll(N, shifts=-1, dims=0),
        A,
        C,
        torch.tensor(1.231),
        torch.tensor(2.108),
        torch.tensor(-3.142),
    ).unsqueeze(2)

    coords = torch.cat(
        [coords, o_coords.view(-1, 1, 3)],
        dim=1,
    )

    return coords


def dist_mat_to_coords(d: torch.Tensor, k: int = 3) -> CoordTensor:
    """
    Computes a set of coordinates that satisfy the (valid) Euclidean
    distance matrix ``d``.

    See: https://math.stackexchange.com/questions/156161/finding-the-coordinates-of-points-from-distance-matrix

    :param d: Valid euclidean distance matrix
    :type d: torch.Tensor
    :param k: Number of dimensions of coordinate output, defaults to ``3``.
    :type k: int, optional
    :return: Set of coordinates that satisfy the distance matrix ``d``.
    :rtype: graphein.protein.tensor.types.CoordTensor
    """
    D_ij_2 = torch.pow(d, 2)
    D_1i_2 = torch.pow(d[0, :], 2).unsqueeze(1)
    D_1j_2 = torch.pow(d[:, 0], 2).unsqueeze(0)
    Mij = (D_1i_2 + D_1j_2 - D_ij_2) / 2

    # try:
    U, S, _ = torch.linalg.svd(Mij)
    # except Exception:  # torch.svd may have convergence issues for GPU and CPU.
    #    U, S, _ = torch.linalg.svd(
    #        Mij + 1e-4 * Mij.mean() * torch.rand_like(Mij)
    #    )
    X = U * torch.sqrt(S)
    X = X[:, :k]
    return X


def build_sidechain(
    bb: AtomTensor,
    residue_types: List[str],
    torsion_angles: TorsionTensor,
) -> AtomTensor:  # sourcery skip: assign-if-exp, lift-duplicated-conditional
    if torsion_angles.shape[-1] == 8:
        torsion_angles = torsion_to_rad(torsion_angles)
        torsion_angles = torch.stack(torsion_angles, dim=1)
    # Initialise new atom tensor [L x 37 x 3] and fill it with backbone atoms.
    # pos = torch.ones_like(x.atom_tensor) * 1e-5
    pos = torch.ones((bb.shape[0], 37, 3), device=bb.device) * 1e-5

    if bb.shape[1] == 5:
        pos[:, :5, :] = bb
        cb_skip = True
    elif bb.shape[1] == 4:
        pos[:, :4, :] = bb
        cb_skip = False

    # Get the residues in the batch (list of 3-letter AA codes)
    # batch_seq = list(chain.from_iterable(x.node_id))
    # nodes = [n.split(":")[1] for n in batch_seq]

    # Iterate over residues
    # for i, res in enumerate(nodes):
    for i, res in enumerate(residue_types):
        start = i == 0
        next_res = pos[i, :] if i == len(residue_types) - 1 else None

        build_vals = _get_build_params(res)
        last_torsion = None
        # Iterate over atoms
        for j, (bond, pbond, angle, torsion, atom_name) in enumerate(
            build_vals
        ):
            # Skip CB as we place it in BB construction
            if atom_name[-1] == "CB" and cb_skip:
                continue
            if start and next_res is not None:
                a, b, c = pos[i + 1, 0], pos[i, 1], pos[i, 2]  # N+, CA, C
            elif start:
                a, b, c = pos[i - 1, 2], pos[i, 0], pos[i, 1]  # C-, N, CA
            else:
                idxs = [ATOM_NUMBERING[atom] for atom in atom_name[:-1]]
                a, b, c = (
                    pos[i, idxs[0]],
                    pos[i, idxs[1]],
                    pos[i, idxs[2]],
                )  # N, CA, C

            # Select correct torsion angle
            if type(torsion) is str and torsion == "p":
                if j == 5 and res == "ARG":
                    torsion = torch.tensor(0.0)
                else:
                    torsion = torsion_angles[i, j - 1]
            elif (
                type(torsion) is str
                and torsion == "i"
                and last_torsion is not None
            ):
                torsion = last_torsion - np.pi

            # Get coordinates of new point
            new_pt = sn_nerf(a, b, c, bond, angle, torsion, pbond)
            # Assign new point to atom tensor
            pos[i, ATOM_NUMBERING[atom_name[-1]]] = new_pt
            last_torsion = torsion

    return pos


def _get_build_params(
    res: str,
) -> Iterable[
    Tuple[
        torch.Tensor,
        torch.Tensor,
        torch.Tensor,
        Union[torch.Tensor, str],
        List[str],
    ]
]:
    """
    Returns iterable of sidechain construction params for a residue.

    Derived from OpenFold:
    https://github.com/aqlaboratory/openfold/blob/main/openfold/np/residue_constants.py

    :param res: Residue name (3-letter code)
    :type res: str
    :return: Iterable of sidechain construction params
    :rtype: Iterable[Tuple[torch.Tensor, torch.Tensor, torch.Tensor, Union[torch.Tensor, str], List[str]]]
    """
    build_params = SC_BUILD_INFO[res]
    bond_vals = [
        torch.tensor(b, dtype=torch.float32)
        for b in build_params["bonds-vals"]
    ]
    pbond_vals = [
        torch.tensor(BB_BUILD_INFO["BONDLENS"]["n-ca"], dtype=torch.float32)
    ] + [
        torch.tensor(b, dtype=torch.float32)
        for b in build_params["bonds-vals"]
    ][
        :-1
    ]
    angle_vals = [
        torch.tensor(a, dtype=torch.float32)
        for a in build_params["angles-vals"]
    ]
    torsion_vals = [
        torch.tensor(t, dtype=torch.float32) if t not in ["p", "i"] else t
        for t in build_params["torsion-vals"]
    ]
    atom_names = [t.split("-") for t in build_params["torsion-names"]]

    return zip(bond_vals, pbond_vals, angle_vals, torsion_vals, atom_names)



"""Utility Functions for working with RNA Secondary Structure Graphs."""

import os
import shutil
from typing import Tuple

import wget

# %%
# Graphein
# Author: Arian Jamasb <arian@jamasb.io>
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein
# This submodule is heavily inspired by: https://github.com/emalgorithm/rna-design/blob/aec77a18abe4850958d6736ec185a6f8cbfdf20c/src/util.py#L9
from loguru import logger as log

BP_RNA_1M_DB_URL: str = (
    "http://bprna.cgrb.oregonstate.edu/bpRNA_1m/dbnFiles.zip"
)
BP_RNA_1M_90_DB_URL: str = "http://bprna.cgrb.oregonstate.edu/bpRNA_1m_90.zip"


def download_bp_rna_1m(out_path: str = "/tmp/"):
    """Downloads BP RNA Dotbrackets.

    :param out_path: _description_, defaults to "/tmp/"
    :type out_path: str, optional
    """
    if os.path.isdir(f"{out_path}/bp_rna_1m"):
        log.info(
            f"Skipping download. bp_rna_1m already downloaded to {out_path}/bp_rna_1m."
        )
    else:
        log.info(f"Downloading bp_rna_1m to {out_path}/bp_rna_1m...")
        fname = f"{out_path}/bp_rna_1m"
        wget.download(BP_RNA_1M_DB_URL, out=f"{fname}.zip")
        log.info("Unzipping bp_rna_1m...")
        shutil.unpack_archive(
            f"{fname}.zip", extract_dir=f"{out_path}/bp_rna_1m"
        )


def read_dbn_file(path: str) -> Tuple[str, str]:
    """
    Read a dotbracket file to sequence and dotbracket strings.

    :param path: Path to the dotbracket file in ``.dbn`` format.
    :type path: str
    :return: Tuple of sequence and dotbracket strings.
    :rtype: Tuple[str, str]
    """
    with open(path, "r") as f:
        lines = f.readlines()
    print(lines)
    sequence = lines[-2]
    dotbracket = lines[-1]
    return sequence, dotbracket


if __name__ == "__main__":
    # download_bp_rna_1m(".")
    seq, db = read_dbn_file("./bp_rna_1m/dbnFiles/bpRNA_CRW_23156.dbn")
    print(seq)
    print(db)



"""Utilities for loading FoldComp databases for deep learning."""

# Graphein
# Author: Arian Jamasb <arian@jamasb.io>
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein

import asyncio
import os
import random
import shutil
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Literal,
    Optional,
    Union,
)

import numpy as np
import torch
from loguru import logger as log
from sklearn.model_selection import train_test_split
from torch_geometric import transforms as T
from torch_geometric.data import Data, Dataset
from torch_geometric.loader import DataLoader
from tqdm import tqdm

from graphein.protein.resi_atoms import (
    ATOM_NUMBERING,
    STANDARD_AMINO_ACID_MAPPING_1_TO_3,
    STANDARD_AMINO_ACIDS,
)
from graphein.protein.tensor import Protein
from graphein.utils.dependencies import import_message

try:
    import foldcomp
except ImportError:
    message = import_message(
        "graphein.ml.datasets.foldcomp", "foldcomp", None, True, extras=True
    )
    log.warning(message)

try:
    import lightning as L
except ImportError:
    message = import_message(
        "graphein.ml.datasets.foldcomp", "lightning", "conda-forge", True
    )
    log.warning(message)

try:
    import nest_asyncio

    nest_asyncio.apply()
except ImportError:
    message = import_message(
        "graphein.ml.datasets.foldcomp", "nest_asyncio", None, True
    )
    message += "You can safely ignore this message if you are not working in a Jupyter Notebook."
    log.warning(message)

FOLDCOMP_DATABASE_TYPES: List[str] = [
    "afdb_swissprot_v4",  # AlphaFoldDB Swiss-Prot
    "afdb_uniprot_v4",  # AlphaFoldDB Uniprot
    "afdb_rep_v4",  # AlphaFoldDB Cluster Representatives
    "afdb_rep_dark_v4",  # AlphaFoldDB Cluster Representatives (Dark Clusters)
    "esmatlas",  # ESMAtlas full (v0 + v2023_02)
    "esmatlas_v2023_02",  # ESMAtlas v2023_02
    "highquality_clust30",  # ESMAtlas high-quality
]
"""
Currently supported FoldComp databases. See:
https://github.com/steineggerlab/foldcomp
"""


GraphTransform = Callable[[Union[Data, Protein]], Union[Data, Protein]]


ATOM_MAP = {
    "MET": ["N", "CA", "C", "O", "CB", "CG", "SD", "CE"],
    "ILE": ["N", "CA", "C", "O", "CB", "CG1", "CG2", "CD1"],
    "LEU": ["N", "CA", "C", "O", "CB", "CG", "CD1", "CD2"],
    "ALA": ["N", "CA", "C", "O", "CB"],
    "ASN": ["N", "CA", "C", "O", "CB", "CG", "OD1", "ND2"],
    "PRO": ["N", "CA", "C", "O", "CB", "CG", "CD"],
    "ARG": ["N", "CA", "C", "O", "CB", "CG", "CD", "NE", "CZ", "NH1", "NH2"],
    "HIS": ["N", "CA", "C", "O", "CB", "CG", "ND1", "CD2", "CE1", "NE2"],
    "GLU": ["N", "CA", "C", "O", "CB", "CG", "CD", "OE1", "OE2"],
    "TYR": [
        "N",
        "CA",
        "C",
        "O",
        "CB",
        "CG",
        "CD1",
        "CD2",
        "CE1",
        "CE2",
        "CZ",
        "OH",
    ],
    "VAL": ["N", "CA", "C", "O", "CB", "CG1", "CG2"],
    "LYS": ["N", "CA", "C", "O", "CB", "CG", "CD", "CE", "NZ"],
    "THR": ["N", "CA", "C", "O", "CB", "OG1", "CG2"],
    "PHE": ["N", "CA", "C", "O", "CB", "CG", "CD1", "CD2", "CE1", "CE2", "CZ"],
    "GLY": ["N", "CA", "C", "O"],
    "SER": ["N", "CA", "C", "O", "CB", "OG"],
    "GLN": ["N", "CA", "C", "O", "CB", "CG", "CD", "OE1", "NE2"],
    "ASP": ["N", "CA", "C", "O", "CB", "CG", "OD1", "OD2"],
    "CYS": ["N", "CA", "C", "O", "CB", "SG"],
    "TRP": [
        "N",
        "CA",
        "C",
        "O",
        "CB",
        "CG",
        "CD1",
        "CD2",
        "NE1",
        "CE2",
        "CE3",
        "CZ2",
        "CZ3",
        "CH2",
    ],
}


class FoldCompDataset(Dataset):
    def __init__(
        self,
        root: str,
        database: Literal[
            "afdb_swissprot_v4",
            "afdb_uniprot_v4",
            "afdb_rep_v4",
            "afdb_rep_dark_v4",
            "esmatlas",
            "esmatlas_v2023_02",
            "highquality_clust30",
        ],
        ids: Optional[List[str]] = None,
        exclude_ids: Optional[List[str]] = None,
        fraction: float = 1.0,
        use_graphein: bool = True,
        transform: Optional[T.BaseTransform] = None,
    ):
        """Dataset class for FoldComp databases.

        :param root: Directory where the dataset should be saved.
        :type root: str
        :param database: Name of the database. See:
            :const:`FOLDCOMP_DATABASE_TYPES`.
        :type database: Literal
        :param ids: List of protein IDs to include in the dataset. If ``None``,
            all proteins are included. Default is ``None``.
        :type ids: Optional[List[str]]
        :param exclude_ids: List of protein IDs to exclude from the dataset.
            If ``None``, all proteins are included. Default is ``None``.
        :type exclude_ids: Optional[List[str]]
        :param fraction: Fraction of database to use, defaults to ``1.0``.
        :type fraction: float, optional
        :param use_graphein: Whether or not to use Graphein's ``Protein``
            objects or to use standard PyG ``Data``, defaults to ``True``.
        :type use_graphein: bool, optional
        :param transform: List of functions/transform that take in a
            ``Data``/``Protein`` object and return a transformed version.
            The data object will be transformed before every access.
            (default: ``None``).
        :type transform: Optional[List[GraphTransform]]
        """
        self.root = Path(root).resolve()
        if not os.path.exists(self.root):
            os.makedirs(self.root)
        self.database = database
        self.ids = ids
        self.exclude_ids = exclude_ids
        self.fraction = fraction
        self.use_graphein = use_graphein
        self.transform = transform

        self._database_files = [
            f"{self.database}",
            f"{self.database}.dbtype",
            f"{self.database}.index",
            f"{self.database}.lookup",
            f"{self.database}.source",
        ]
        self._get_indices()

        super().__init__(
            root=self.root,
            transform=self.transform,
            pre_transform=None,  # type: ignore
        )

    @property
    def raw_file_names(self) -> List[str]:
        """Return a list of raw file names expected to be found in the database.

        Returns empty string as we don't expect any raw files.

        :return: List of raw file names
        :rtype: List[str]
        """
        return [""]

    @property
    def processed_file_names(self):
        """Returns a token to force self.process() to be called."""
        return ["_"]

    def download(self):
        """Downloads foldcomp database if not already downloaded."""

        if not all(
            os.path.exists(self.root / f) for f in self._database_files
        ):
            log.info(f"Downloading FoldComp dataset {self.database}...")
            curr_dir = os.getcwd()
            os.chdir(self.root)
            try:
                foldcomp.setup(self.database)
            except RuntimeError:
                _ = self.async_setup()
                asyncio.run(_)
            os.chdir(curr_dir)
            log.info("Download complete.")
        else:
            log.info(f"FoldComp database already downloaded: {self.root}.")

    async def async_setup(self):
        """Asynchronous setup of foldcomp database (for jupyter support)."""
        await foldcomp.setup_async(self.database)

    def _get_indices(self):
        """Get indices for the dataset."""
        # Read in look up file
        LOOKUP_FILE = Path(self.root) / f"{self.database}.lookup"
        if not os.path.exists(LOOKUP_FILE):
            self.download()
        with open(LOOKUP_FILE, "r") as f:
            accessions = f.readlines()
        # Extract accessions
        accessions = [x.strip().split("\t")[1] for x in tqdm(accessions)]
        # Get indices
        if self.ids is not None:
            accessions = (
                self.ids
            )  # [acc for acc in accessions if acc in tqdm(self.ids)]
        # Exclude indices
        if self.exclude_ids is not None:
            log.info(f"Excluding {len(self.exclude_ids)} chains...")
            accessions = [
                acc for acc in tqdm(accessions) if acc not in self.exclude_ids
            ]
        # Sub sample
        log.info(f"Sampling fraction: {self.fraction}...")
        accessions = random.sample(
            accessions, int(len(accessions) * self.fraction)
        )
        self.ids = accessions
        log.info("Creating index...")
        indices = dict(enumerate(accessions))
        self.idx_to_protein = indices
        self.protein_to_idx = {v: k for k, v in indices.items()}
        log.info(f"Dataset contains {len(self.protein_to_idx)} chains.")

    def process(self):
        """Initialises the database."""
        # Open the database
        log.info("Opening database...")
        if self.ids is not None:
            self.db = foldcomp.open(
                self.root / self.database, ids=self.ids, decompress=False
            )  # type: ignore
        else:
            self.db = foldcomp.open(self.root / self.database, decompress=False)  # type: ignore

    @staticmethod
    def fc_to_pyg(data: Dict[str, Any], name: Optional[str] = None) -> Protein:
        # Map sequence to 3-letter codes
        res = [STANDARD_AMINO_ACID_MAPPING_1_TO_3[r] for r in data["residues"]]
        residue_type = torch.tensor(
            [STANDARD_AMINO_ACIDS.index(res) for res in data["residues"]],
        )
        n_res = len(res)

        # Get residue numbers
        res_num = np.arange(n_res)

        # Get list of atom types
        atom_types = []
        atom_counts = []
        for r in res:
            atom_types += ATOM_MAP[r]
            atom_counts.append(len(ATOM_MAP[r]))
        atom_types += ["OXT"]
        atom_counts[-1] += 1

        # Get atom indices
        atom_idx = np.array([ATOM_NUMBERING[atm] for atm in atom_types])

        # Initialize coordinates
        coords = np.ones((n_res, 37, 3)) * 1e-5

        res_idx = np.repeat(res_num, atom_counts)
        coords[res_idx, atom_idx, :] = np.array(data["coordinates"])
        b_factor = np.array(data["b_factors"]) / 100

        return Protein(
            coords=torch.from_numpy(coords).float(),
            residues=res,
            residue_id=[f"A:{m}:{str(n)}" for m, n in zip(res, res_num)],
            chains=torch.zeros(n_res),
            residue_type=residue_type.long(),
            b_factor=torch.from_numpy(b_factor).float(),
            id=name,
            x=torch.zeros(n_res),
            seq_pos=torch.from_numpy(res_num).unsqueeze(-1),
        )

    def len(self) -> int:
        """Returns length of the dataset"""
        return len(self.protein_to_idx)

    def get(self, idx) -> Union[Data, Protein]:
        """Retrieves a protein from the dataset. Can idx on either the protein
        ID or its index."""
        if isinstance(idx, str):
            idx = self.protein_to_idx[idx]

        name = self.idx_to_protein[idx]
        data = foldcomp.get_data(self.db[idx])  # type: ignore
        return self.fc_to_pyg(data, name)


class FoldCompLightningDataModule(L.LightningDataModule):
    def __init__(
        self,
        data_dir,
        database: str,
        batch_size: int,
        use_graphein: bool = True,
        train_split: Optional[Union[List[str], float]] = None,
        val_split: Optional[Union[List[str], float]] = None,
        test_split: Optional[Union[List[str], float]] = None,
        transform: Optional[Iterable[Callable]] = None,
        num_workers: int = 4,
        pin_memory: bool = True,
    ) -> None:
        """Creates a PyTorch Lightning DataModule for FoldComp datasets.

        :param data_dir: Path to dataset directory.
        :type data_dir: str
        :param database: Name of the FoldComp database to use. See:
            :const:`FOLDCOMP_DATABASE_TYPES`.
        :type database: str
        :param batch_size: Batch size to use.
        :type batch_size: int
        :param use_graphein: Whether to use Graphein ``Protein`` objects or
            PyTorch Geometric ``Data`` objects, defaults to ``True``.
        :type use_graphein: bool, optional
        :param train_split: List of IDs or a float specifying fraction of  the
            dataset to use for training, defaults to ``None`` (whole dataset).
        :type train_split: Optional[Union[List[str], float]], optional
        :param val_split: List of IDs or a float specifying fraction of  the
            dataset to use for training, defaults to ``None`` (whole dataset).
        :type val_split: Optional[Union[List[str], float]], optional
        :param test_split: List of IDs or a float specifying fraction of  the
            dataset to use for training, defaults to ``None`` (whole dataset).
        :type test_split: Optional[Union[List[str], float]], optional
        :param transform: List of functions/transform that take in a
            ``Data``/``Protein`` object and return a transformed version.
            The data object will be transformed before every access.
            (default: ``None``).
        :type transform: Optional[Iterable[Callable]]
        :param num_workers: Number of workers to use for data loading, defaults
            to ``4``.
        :type num_workers: int, optional
        :param pin_memory: Whether to pin memory for data loading, defaults to
            ``True``
        :type pin_memory: bool, optional
        """
        super().__init__()
        self.data_dir = data_dir
        self.database = database
        self.batch_size = batch_size
        self.use_graphein = use_graphein
        self.train_split = train_split
        self.val_split = val_split
        self.test_split = test_split
        self.transform = (
            self._compose_transforms(transform)
            if transform is not None
            else None
        )

        if (
            isinstance(train_split, float)
            and isinstance(val_split, float)
            and isinstance(test_split, float)
        ):
            self._split_data(train_split, val_split, test_split)
            log.info(
                f"Split data into train ({train_split}, {len(self.train_split)}), val ({val_split}, {len(self.val_split)}) and test ({test_split}, {len(self.test_split)}) sets."
            )

        self.num_workers = num_workers
        self.pin_memory = pin_memory

    def _compose_transforms(self, transforms: Iterable[Callable]) -> T.Compose:
        try:
            return T.Compose(list(transforms.values()))
        except Exception:
            return T.Compose(transforms)

    def setup(self, stage: Optional[str] = None):
        if stage == "fit" or stage is None:
            self.train_dataset()
        elif stage == "validate":
            self.val_dataset()
        elif stage == "test":
            self.test_dataset()

    def _get_indices(self):
        """Loads the whole database to extract the indices."""
        log.info("Getting indices (loading complete database)...")
        ds = FoldCompDataset(
            self.data_dir,
            self.database,
            ids=None,
            use_graphein=self.use_graphein,
        )
        self.ids = ds.ids
        ds.db.close()

    def _split_data(
        self, train_split: float, val_split: float, test_split: float
    ):
        """Split the database into non-overlapping train, validation and test"""
        if not hasattr(self, "ids"):
            self._get_indices()
            train, test = train_test_split(self.ids, test_size=1 - train_split)

            size = test_split / (test_split + val_split)
            val, test = train_test_split(test, test_size=size)
            self.train_split = train
            self.val_split = val
            self.test_split = test

    def _create_dataset(self, split: str) -> FoldCompDataset:
        """Initialises a FoldCompDataset for a given split.

        :param split: Split to initialise. Must be one of ``train``, ``val``, or
            ``test``.
        :type split: str
        :raises ValueError: If the split is invalid.
        :return: FoldCompDataset for the given split.
        :rtype: FoldCompDataset
        """
        data_split = getattr(self, f"{split}_split")
        return FoldCompDataset(
            self.data_dir,
            self.database,
            ids=data_split,
            use_graphein=self.use_graphein,
            transform=self.transform,
        )

    def train_dataset(self):
        """Initialises the training dataset."""
        log.info("Creating training dataset...")
        self.train_ds = self._create_dataset("train")

    def val_dataset(self):
        """Initialises the validation dataset."""
        log.info("Creating validation dataset...")
        self.val_ds = self._create_dataset("val")

    def test_dataset(self):
        """Initialises the test dataset."""
        log.info("Creating test dataset...")
        self.test_ds = self._create_dataset("test")

    def train_dataloader(self) -> DataLoader:
        """Returns a training dataloader.

        :return: Training dataloader.
        :rtype: DataLoader
        """
        return DataLoader(
            self.train_ds,
            batch_size=self.batch_size,
            shuffle=True,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
        )

    def val_dataloader(self) -> DataLoader:
        """Returns a validation dataloader.

        :return: Validation dataloader.
        :rtype: DataLoader
        """
        return DataLoader(
            self.val_ds,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
        )

    def test_dataloader(self) -> DataLoader:
        """Returns a test dataloader.

        :return: Test dataloader.
        :rtype: DataLoader
        """
        return DataLoader(
            self.test_ds,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
        )



"""Contains utilities for plotting PPI NetworkX graphs."""

# %%
# Graphein
# Author: Arian Jamasb <arian@jamasb.io>
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein
from typing import List, Optional

import networkx as nx
import plotly.express as px
import plotly.graph_objects as go
from matplotlib.colors import to_rgb


def plot_ppi_graph(
    g: nx.Graph,
    colour_edges_by: str = "kind",
    with_labels: bool = True,
    **kwargs,
):
    """Plots a Protein-Protein Interaction Graph. Colours edges by kind.

    :param g: NetworkX graph of PPI network.
    :type g: nx.Graph
    :param colour_edges_by: Colour edges by this attribute. Currently, only supports 'kind', which colours edges by the source database, by default "kind"
    :param with_labels: Whether to show labels on nodes. Defaults to True.
    :type with_labels: bool, optional
    """
    if colour_edges_by == "kind":
        edge_colors = [
            (
                "r"
                if g[u][v]["kind"] == {"string"}
                else "b" if g[u][v]["kind"] == {"biogrid"} else "y"
            )
            for u, v in g.edges()
        ]
    else:
        raise ValueError(
            f"Edge colouring scheme: {colour_edges_by} not supported. Please use 'kind'"
        )
    nx.draw(g, with_labels=with_labels, edge_color=edge_colors, **kwargs)


def plotly_ppi_graph(
    g: nx.Graph,
    layout: nx.layout = nx.layout.circular_layout,
    title: Optional[str] = None,
    show_labels: bool = False,
    node_size_multiplier: float = 5.0,
    node_colourscale: str = "Viridis",
    edge_colours: Optional[List[str]] = None,
    edge_opacity: float = 0.5,
    height: int = 500,
    width: int = 500,
):
    """Plots a PPI graph.

    :param g: PPI graph
    :type g: nx.Graph
    :param layout: Layout algorithm to use. Default is circular_layout.
    :type layout: nx.layout
    :param title: Title of the graph. Default is None.
    :type title: str, optional
    :param show_labels: If True, shows labels on nodes. Default is False.
    :type show_labels: bool
    :param node_size_multiplier: Multiplier for node size. Default is 5.0.
    :type node_size_multiplier: float
    :param node_colourscale: Colour scale to use for node colours. Default is "Viridis". Options:
        'Greys' | 'YlGnBu' | 'Greens' | 'YlOrRd' | 'Bluered' | 'RdBu' |
        'Reds' | 'Blues' | 'Picnic' | 'Rainbow' | 'Portland' | 'Jet' |
        'Hot' | 'Blackbody' | 'Earth' | 'Electric' | 'Viridis' |
    :type node_colourscale: str
    :param edge_colours: List of colours (hexcode) to use for edges. Default is None (px.colours.qualitative.T10).
    :type edge_colours: List[str], optional
    :param edge_opacity: Opacity of edges. Default is 0.5.
    :type edge_opacity: float
    :param height: Height of the plot. Default is 500.
    :type height: int
    :param width: Width of the plot. Default is 500.
    :type width: int
    :return: Plotly figure of PPI Network
    :rtype: go.Figure
    """
    if edge_colours is None:
        edge_colours = px.colors.qualitative.T10
    edge_colours = [
        f"rgba{tuple(list(to_rgb(c)) + [edge_opacity])}" for c in edge_colours
    ]

    # Set positions
    nx.set_node_attributes(g, layout(g), "pos")

    # Get node and edge traces
    node_trace = get_node_trace(g, node_size_multiplier, node_colourscale)
    edge_trace = get_edge_trace(g, edge_colours)
    traces = [node_trace] + edge_trace

    # Get node labels if using them.
    if show_labels:
        text_trace = go.Scatter(
            x=node_trace["x"],
            y=node_trace["y"],
            mode="text",
            text=list(g.nodes()),
            textposition="bottom center",
            hoverinfo="text",
        )
        traces.append(text_trace)

    # Assemble plot from traces
    return go.Figure(
        data=traces,
        layout=go.Layout(
            title=title,
            showlegend=False,
            width=width,
            height=height,
            hovermode="closest",
            margin=dict(b=20, l=5, r=5, t=40),
            xaxis=dict(showgrid=False, zeroline=False, showticklabels=False),
            yaxis=dict(showgrid=False, zeroline=False, showticklabels=False),
        ),
    )


def get_node_trace(
    g: nx.Graph, node_size_multiplier: float, node_colourscale: str = "Viridis"
) -> go.Scatter:
    """Produces the node trace for the plotly plot.

    :param g: PPI graph with ['pos'] added to the nodes (eg via nx.layout function)
    :type g: nx.Graph
    :param node_size_multiplier: Multiplier for node size. Default is 5.0.
    :type node_size_multiplier: float
    :param node_colourscale: Colourscale to use for the nodes, defaults to "Viridis"
    :type node_colourscale: str, optional
    :return: Node trace for plotly plot
    :rtype: go.Scatter
    """
    node_x = []
    node_y = []
    node_size = []
    for n in g.nodes():
        x, y = g.nodes[n]["pos"]
        node_x.append(x)
        node_y.append(y)
        node_size.append(g.degree(n) * node_size_multiplier)

    node_trace = go.Scatter(
        x=node_x,
        y=node_y,
        mode="markers",
        hoverinfo="text",
        marker=dict(
            showscale=False,
            colorscale=node_colourscale,
            reversescale=True,
            color=[],
            size=node_size,
            colorbar=dict(
                thickness=15,
                title="Node Connections",
                xanchor="left",
            ),
            line_width=2,
        ),
    )
    node_text = list(g.nodes())
    node_trace.marker.color = node_size
    node_trace.text = node_text
    return node_trace


def get_edge_trace(
    g: nx.Graph,
    edge_colours: Optional[List[str]] = None,
) -> List[go.Scatter]:
    """Gets edge traces from PPI graph. Returns a list of traces enabling edge colours to be set individually.

    :param g: _description_
    :type g: nx.Graph
    :return: _description_
    :rtype: List[go.Scatter]
    """

    if edge_colours is None:
        edge_colours = ["red", "blue", "yellow"]
    traces = []
    for u, v, d in g.edges(data=True):
        # Get positions
        x0, y0 = g.nodes[u]["pos"]
        x1, y1 = g.nodes[v]["pos"]
        # Assign colour
        if d["kind"] == {"string"}:
            colour = edge_colours[0]
        elif d["kind"] == {"biogrid"}:
            colour = edge_colours[1]
        else:
            colour = edge_colours[2]

        edge_trace = go.Scatter(
            line=dict(width=2, color=colour),
            hoverinfo="text",
            x=(x0, x1),
            y=(y0, y1),
            mode="lines",
            text=[
                " / ".join(list(edge_type)) for edge_type in g[u][v]["kind"]
            ],
        )
        traces.append(edge_trace)
    return traces


if __name__ == "__main__":
    from functools import partial

    from graphein.ppi.config import PPIGraphConfig
    from graphein.ppi.edges import add_biogrid_edges, add_string_edges
    from graphein.ppi.graphs import compute_ppi_graph

    config = PPIGraphConfig()

    protein_list = [
        "CDC42",
        "CDK1",
        "KIF23",
        "PLK1",
        "RAC2",
        "RACGAP1",
        "RHOA",
        "RHOB",
    ]

    g = compute_ppi_graph(
        protein_list=protein_list,
        edge_construction_funcs=[
            partial(add_string_edges),
            partial(add_biogrid_edges),
        ],
    )

    plot_ppi_graph(g)
    plotly_ppi_graph(g)

# %%



"""Provides functions for extracting subgraphs from protein graphs."""

# Graphein
# Author: Arian Jamasb <arian@jamasb.io>
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein

from typing import Dict, List, Optional, Tuple, Union

import networkx as nx
import numpy as np
from loguru import logger as log

from graphein.protein.edges.distance import compute_distmat
from graphein.protein.utils import ProteinGraphConfigurationError


def extract_subgraph_from_node_list(
    g,
    node_list: Optional[List[str]],
    filter_dataframe: bool = True,
    update_coords: bool = True,
    recompute_distmat: bool = False,
    inverse: bool = False,
    return_node_list: bool = False,
) -> Union[nx.Graph, List[str]]:
    """Extracts a subgraph from a graph based on a list of nodes.

    :param g: The graph to extract the subgraph from.
    :type g: nx.Graph
    :param node_list: The list of nodes to extract.
    :type node_list: List[str]
    :param filter_dataframe: Whether to filter the ``pdb_df`` DataFrame of the
        graph. Defaults to ``True``.
    :type filter_dataframe: bool
    :param update_coords: Whether to update the coordinates of the graph.
        Defaults to ``True``.
    :type update_coords: bool
    :param recompute_distmat: Whether to recompute the distance matrix of the
        graph. Defaults to ``False``.
    :type recompute_distmat: bool
    :param inverse: Whether to inverse the selection. Defaults to ``False``.
    :type inverse: bool
    :return: The subgraph or node list if ``return_node_list=True``.
    :rtype: Union[nx.Graph, List[str]]
    """
    if node_list:
        # Get all nodes not in nodelist if inversing the selection
        if inverse:
            node_list = [n for n in g.nodes() if n not in node_list]

        # If we are just returning the node list, return it here before
        # subgraphing.
        if return_node_list:
            return node_list

        log.debug(f"Creating subgraph from nodes: {node_list}.")
        # Create a subgraph from the node list.
        g = g.subgraph(node_list).copy()
        # Filter the PDB DF accordingly
        if filter_dataframe:
            g.graph["pdb_df"] = g.graph["pdb_df"].loc[
                g.graph["pdb_df"]["node_id"].isin(node_list)
            ]
        if update_coords:
            g.graph["coords"] = np.array(
                [d["coords"] for _, d in g.nodes(data=True)]
            )
        if recompute_distmat:
            if not filter_dataframe:
                log.warning("Recomputing distmat without filtering dataframe.")
            g.graph["dist_mat"] = compute_distmat(g.graph["pdb_df"])
        # Reset numbering for edge funcs
        g.graph["pdb_df"] = g.graph["pdb_df"].reset_index(drop=True)

    return node_list if return_node_list else g


def extract_subgraph_from_point(
    g: nx.Graph,
    centre_point: Union[np.ndarray, Tuple[float, float, float]],
    radius: float,
    filter_dataframe: bool = True,
    update_coords: bool = True,
    recompute_distmat: bool = False,
    inverse: bool = False,
    return_node_list: bool = False,
) -> Union[nx.Graph, List[str]]:
    """Extracts a subgraph from a graph based on a centre point and radius.

    :param g: The graph to extract the subgraph from.
    :type g: nx.Graph
    :param centre_point: The centre point of the subgraph.
    :type centre_point: Tuple[float, float, float]
    :param radius: The radius of the subgraph.
    :type radius: float
    :param filter_dataframe: Whether to filter the pdb_df dataframe of the
        graph. Defaults to ``True``.
    :type filter_dataframe: bool
    :param update_coords: Whether to update the coordinates of the graph.
        Defaults to ``True``.
    :type update_coords: bool
    :param recompute_distmat: Whether to recompute the distance matrix of the
        graph. Defaults to ``False``.
    :type recompute_distmat: bool
    :param inverse: Whether to inverse the selection. Defaults to ``False``.
    :type inverse: bool
    :return: The subgraph or node list if ``return_node_list=True``.
    :rtype: Union[nx.Graph, List[str]]
    """
    node_list: List = []

    for n, d in g.nodes(data=True):
        coords = d["coords"]
        dist = np.linalg.norm(coords - centre_point)
        if dist < radius:
            node_list.append(n)

    node_list = list(set(node_list))
    log.debug(
        f"Found {len(node_list)} nodes in the spatial point-radius subgraph."
    )

    return extract_subgraph_from_node_list(
        g,
        node_list,
        filter_dataframe=filter_dataframe,
        inverse=inverse,
        return_node_list=return_node_list,
        recompute_distmat=recompute_distmat,
        update_coords=update_coords,
    )


def extract_subgraph_from_atom_types(
    g: nx.Graph,
    atom_types: List[str],
    filter_dataframe: bool = True,
    update_coords: bool = True,
    recompute_distmat: bool = False,
    inverse: bool = False,
    return_node_list: bool = False,
) -> Union[nx.Graph, List[str]]:
    """Extracts a subgraph from a graph based on a list of atom types.

    :param g: The graph to extract the subgraph from.
    :type g: nx.Graph
    :param atom_types: The list of atom types to extract.
    :type atom_types: List[str]
    :param filter_dataframe: Whether to filter the ``pdb_df`` dataframe of the
        graph. Defaults to ``True``.
    :type filter_dataframe: bool
    :param update_coords: Whether to update the coordinates of the graph.
        Defaults to ``True``.
    :type update_coords: bool
    :param recompute_distmat: Whether to recompute the distance matrix of the
        graph. Defaults to ``False``.
    :type recompute_distmat: bool
    :param inverse: Whether to inverse the selection. Defaults to ``False``.
    :type inverse: bool
    :return: The subgraph or node list if ``return_node_list=True``.
    :rtype: Union[nx.Graph, List[str]]
    """
    node_list: List = [
        n for n, d in g.nodes(data=True) if d["atom_type"] in atom_types
    ]

    node_list = list(set(node_list))
    log.debug(f"Found {len(node_list)} nodes in the atom type subgraph.")

    return extract_subgraph_from_node_list(
        g,
        node_list,
        filter_dataframe=filter_dataframe,
        inverse=inverse,
        return_node_list=return_node_list,
        update_coords=update_coords,
        recompute_distmat=recompute_distmat,
    )


def extract_subgraph_from_residue_types(
    g: nx.Graph,
    residue_types: List[str],
    filter_dataframe: bool = True,
    update_coords: bool = True,
    recompute_distmat: bool = False,
    inverse: bool = False,
    return_node_list: bool = False,
) -> Union[nx.Graph, List[str]]:
    """Extracts a subgraph from a graph based on a list of allowable residue
    types.

    :param g: The graph to extract the subgraph from.
    :type g: nx.Graph
    :param residue_types: List of allowable residue types (3 letter residue
        names). E.g. ``["SER", "GLY", "ALA"]``
    :type residue_types: List[str]
    :param filter_dataframe: Whether to filer the pdb_df of the graph, defaults
        to ``True``
    :type filter_dataframe: bool, optional
    :param update_coords: Whether to update the coordinates of the graph.
        Defaults to ``True``.
    :type update_coords: bool
    :param recompute_distmat: Whether to recompute the distance matrix of the
        graph. Defaults to ``False``.
    :type recompute_distmat: bool
    :param inverse: Whether to inverse the selection. Defaults to ``False``.
    :type inverse: bool
    :return: The subgraph or node list if ``return_node_list=True``.
    :rtype: Union[nx.Graph, List[str]]
    """
    node_list: List = [
        n for n, d in g.nodes(data=True) if d["residue_name"] in residue_types
    ]

    node_list = list(set(node_list))
    log.debug(f"Found {len(node_list)} nodes in the residue type subgraph.")

    return extract_subgraph_from_node_list(
        g,
        node_list,
        filter_dataframe=filter_dataframe,
        inverse=inverse,
        return_node_list=return_node_list,
        update_coords=update_coords,
        recompute_distmat=recompute_distmat,
    )


def extract_subgraph_from_chains(
    g: nx.Graph,
    chains: List[str],
    filter_dataframe: bool = True,
    update_coords: bool = True,
    recompute_distmat: bool = False,
    inverse: bool = False,
    return_node_list: bool = False,
) -> Union[nx.Graph, List[str]]:
    """Extracts a subgraph from a graph based on a chain.

    :param g: The graph to extract the subgraph from.
    :type g: nx.Graph
    :param chain: The chain(s) to extract. E.g. ``["A", "C", "E"]`` or
        ``["A"]``.
    :type chain: List[str]
    :param filter_dataframe: Whether to filter the ``pdb_df`` dataframe of the
        graph. Defaults to ``True``.
    :type filter_dataframe: bool
    :param update_coords: Whether to update the coordinates of the graph.
        Defaults to ``True``.
    :type update_coords: bool
    :param recompute_distmat: Whether to recompute the distance matrix of the
        graph. Defaults to ``False``.
    :type recompute_distmat: bool
    :param inverse: Whether to inverse the selection. Defaults to ``False``.
    :type inverse: bool
    :return: The subgraph or node list if ``return_node_list=True``.
    :rtype: Union[nx.Graph, List[str]]
    """
    node_list: List = [
        n for n, d in g.nodes(data=True) if d["chain_id"] in chains
    ]

    node_list = list(set(node_list))
    log.debug(f"Found {len(node_list)} nodes in the chain subgraph.")
    return extract_subgraph_from_node_list(
        g,
        node_list,
        filter_dataframe=filter_dataframe,
        update_coords=update_coords,
        recompute_distmat=recompute_distmat,
        inverse=inverse,
        return_node_list=return_node_list,
    )


def extract_subgraph_by_sequence_position(
    g: nx.Graph,
    sequence_positions: List[int],
    filter_dataframe: bool = True,
    update_coords: bool = True,
    recompute_distmat: bool = False,
    inverse: bool = False,
    return_node_list: bool = False,
) -> Union[nx.Graph, List[str]]:
    """Extracts a subgraph from a graph based on position in the sequence

    :param g: The graph to extract the subgraph from.
    :type g: nx.Graph
    :param chain: The sequence positions to extract. E.g. ``[1, 2, 3]`` or
        ``[1]``.
    :type chain: List[int]
    :param filter_dataframe: Whether to filter the ``pdb_df`` dataframe of the
        graph. Defaults to ``True``.
    :type filter_dataframe: bool
    :param update_coords: Whether to update the coordinates of the graph.
        Defaults to ``True``.
    :type update_coords: bool
    :param recompute_distmat: Whether to recompute the distance matrix of the
        graph. Defaults to ``False``.
    :type recompute_distmat: bool
    :param inverse: Whether to inverse the selection. Defaults to ``False``.
    :type inverse: bool
    :return: The subgraph or node list if ``return_node_list=True``.
    :rtype: Union[nx.Graph, List[str]]
    """
    node_list: List = [
        n
        for n, d in g.nodes(data=True)
        if d["residue_number"] in sequence_positions
    ]

    node_list = list(set(node_list))
    log.debug(
        f"Found {len(node_list)} nodes in the sequence position subgraph."
    )
    return extract_subgraph_from_node_list(
        g,
        node_list,
        filter_dataframe=filter_dataframe,
        inverse=inverse,
        return_node_list=return_node_list,
        recompute_distmat=recompute_distmat,
        update_coords=update_coords,
    )


def extract_subgraph_by_bond_type(
    g: nx.Graph,
    bond_types: List[str],
    filter_dataframe: bool = True,
    update_coords: bool = True,
    recompute_distmat: bool = False,
    inverse: bool = False,
    return_node_list: bool = False,
) -> Union[nx.Graph, List[str]]:
    """Extracts a subgraph from a graph based on a list of allowable bond types.

    :param g: The graph to extract the subgraph from.
    :type g: nx.Graph
    :param bond_types: List of allowable bond types. E.g.
        ``["hydrogen_bond", "k_nn_3"]``
    :type bond_types: List[str]
    :param filter_dataframe: Whether to filter the ``pdb_df`` of the graph,
        defaults to ``True``.
    :type filter_dataframe: bool, optional
    :param update_coords: Whether to update the coordinates of the graph.
        Defaults to ``True``.
    :type update_coords: bool
    :param recompute_distmat: Whether to recompute the distance matrix of the
        graph. Defaults to ``False``.
    :type recompute_distmat: bool
    :param inverse: Whether to inverse the selection, defaults to ``False``.
    :type inverse: bool, optional
    :param return_node_list: Whether to return the node list, defaults to
        ``False``.
    :type return_node_list: bool, optional
    :return: The subgraph or node list if ``return_node_list=True``.
    :rtype: Union[nx.Graph, List[str]]
    """

    node_list: List = []

    for u, v, d in g.edges(data=True):
        for bond_type in list(d["kind"]):
            if bond_type in bond_types:
                node_list.append(u)
                node_list.append(v)
    node_list = list(set(node_list))
    log.debug(f"Found {len(node_list)} nodes in the bond type subgraph.")

    # Remove bond annotations
    for u, v, d in g.edges(data=True):
        for bond in list(d["kind"]):
            if not inverse:
                if bond not in bond_types:
                    d["kind"].discard(bond)
            elif inverse:
                if bond in bond_types:
                    d["kind"].discard(bond)

    return extract_subgraph_from_node_list(
        g,
        node_list,
        filter_dataframe=filter_dataframe,
        inverse=inverse,
        return_node_list=return_node_list,
        recompute_distmat=recompute_distmat,
        update_coords=update_coords,
    )


def extract_subgraph_from_secondary_structure(
    g: nx.Graph,
    ss_elements: List[str],
    inverse: bool = False,
    filter_dataframe: bool = True,
    recompute_distmat: bool = False,
    update_coords: bool = True,
    return_node_list: bool = False,
) -> Union[nx.Graph, List[str]]:
    """Extracts subgraphs for nodes that have a secondary structure element in
    the list.

    :param g: The graph to extract the subgraph from.
    :type g: nx.Graph
    :param ss_elements: List of secondary structure elements to extract. E.g.
        ``["H", "E"]``
    :type ss_elements: List[str]
    :param inverse: Whether to inverse the selection. Defaults to ``False``.
    :type inverse: bool
    :param filter_dataframe: Whether to filter the ``pdb_df`` of the graph,
        defaults to ``True``.
    :type filter_dataframe: bool, optional
    :param recompute_distmat: Whether to recompute the distance matrix of the
        graph. Defaults to ``False``.
    :type recompute_distmat: bool
    :param update_coords: Whether to update the coordinates of the graph.
        Defaults to ``True``.
    :type update_coords: bool
    :param return_node_list: Whether to return the node list. Defaults to
        ``False``.
    :raises ProteinGraphConfigurationError: If the graph does not contain ss
        features on the nodes
        (``d['ss'] not in d.keys() for _, d in g.nodes(data=True)``).
    :return: The subgraph or node list if ``return_node_list=True``.
    :rtype: Union[nx.Graph, List[str]]
    """

    node_list: List[str] = []

    for n, d in g.nodes(data=True):
        if "ss" not in d.keys():
            raise ProteinGraphConfigurationError(
                f"Secondary structure not defined for all nodes ({n}). \
                    Please ensure you have used \
                        graphein.protein.nodes.features.dssp.secondary_structure\
                            as a graph annotation function."
            )
        if d["ss"] in ss_elements:
            node_list.append(n)

    node_list = list(set(node_list))
    log.debug(
        f"Found {len(node_list)} nodes in the secondary structure subgraph."
    )

    return extract_subgraph_from_node_list(
        g,
        node_list,
        inverse=inverse,
        return_node_list=return_node_list,
        filter_dataframe=filter_dataframe,
        recompute_distmat=recompute_distmat,
        update_coords=update_coords,
    )


def extract_surface_subgraph(
    g: nx.Graph,
    rsa_threshold: float = 0.2,
    inverse: bool = False,
    filter_dataframe: bool = True,
    recompute_distmat: bool = False,
    update_coords: bool = True,
    return_node_list: bool = False,
) -> Union[nx.Graph, List[str]]:
    """Extracts a subgraph based on thresholding the Relative Solvent
    Accessibility (RSA). This can be used for extracting a surface graph.

    :param g: The graph to extract the subgraph from.
    :type g: nx.Graph
    :param rsa_threshold: The threshold to use for the RSA. Defaults to
        ``0.2`` (20%).
    :type rsa_threshold: float
    :param filter_dataframe: Whether to filter the ``pdb_df`` of the graph,
        defaults to ``True``.
    :type filter_dataframe: bool, optional
    :param update_coords: Whether to update the coordinates of the graph.
        Defaults to ``True``.
    :type update_coords: bool
    :param recompute_distmat: Whether to recompute the distance matrix of the
        graph. Defaults to ``False``.
    :type recompute_distmat: bool
    :param inverse: Whether to inverse the selection, defaults to ``False``.
    :type inverse: bool, optional
    :param return_node_list: Whether to return the node list. Defaults to
        ``False``.
    :type return_node_list: bool
    :raises ProteinGraphConfigurationError: If the graph does not contain
        RSA features on the nodes
        (``d['rsa'] not in d.keys() for _, d in g.nodes(data=True)``).
    :return: The subgraph or node list if ``return_node_list=True``.
    :rtype: Union[nx.Graph, List[str]]
    """
    node_list: List[str] = []

    for n, d in g.nodes(data=True):
        if "rsa" not in d.keys():
            raise ProteinGraphConfigurationError(
                f"RSA not defined for all nodes ({n}). Please ensure you have \
                    used graphein.protein.nodes.features.dssp.rsa as a graph \
                        annotation function."
            )
        if d["rsa"] >= rsa_threshold:
            node_list.append(n)

    node_list = list(set(node_list))
    log.debug(f"Found {len(node_list)} nodes in the surface subgraph.")

    return extract_subgraph_from_node_list(
        g,
        node_list,
        inverse=inverse,
        return_node_list=return_node_list,
        filter_dataframe=filter_dataframe,
        recompute_distmat=recompute_distmat,
        update_coords=update_coords,
    )


def extract_k_hop_subgraph(
    g: nx.Graph,
    central_node: str,
    k: int,
    k_only: bool = False,
    filter_dataframe: bool = True,
    update_coords: bool = True,
    recompute_distmat: bool = False,
    inverse: bool = False,
    return_node_list: bool = False,
) -> Union[nx.Graph, List[str]]:
    """Extracts a k-hop subgraph.

    :param g: The graph to extract the subgraph from.
    :type g: nx.Graph
    :param central_node: The central node to extract the subgraph from.
        E.g. ``"A:SER:12"``.
    :type central_node: str
    :param k: The number of hops to extract.
    :type k: int
    :param k_only: Whether to only extract the exact k-hop subgraph
        (e.g. include 2-hop neighbours in 5-hop graph). Defaults to ``False``.
    :type k_only: bool
    :param filter_dataframe: Whether to filter the ``pdb_df`` of the graph,
        defaults to ``True``.
    :type filter_dataframe: bool, optional
    :param update_coords: Whether to update the coordinates of the graph.
        Defaults to ``True``.
    :type update_coords: bool
    :param recompute_distmat: Whether to recompute the distance matrix of the
        graph. Defaults to ``False``.
    :type recompute_distmat: bool
    :param inverse: Whether to inverse the selection, defaults to ``False``.
    :type inverse: bool, optional
    :param return_node_list: Whether to return the node list. Defaults to
        ``False``.
    :type return_node_list: bool
    :return: The subgraph or node list if ``return_node_list=True``.
    :rtype: Union[nx.Graph, List[str]]
    """
    neighbours: Dict[int, Union[List[str], set]] = {0: [central_node]}

    for i in range(1, k + 1):
        neighbours[i] = set()
        for node in neighbours[i - 1]:
            neighbours[i].update(g.neighbors(node))
        neighbours[i] = list(set(neighbours[i]))

    if k_only:
        node_list = neighbours[k]
    else:
        node_list = list(
            {value for values in neighbours.values() for value in values}
        )

    log.debug(f"Found {len(node_list)} nodes in the k-hop subgraph.")

    return extract_subgraph_from_node_list(
        g,
        node_list,
        filter_dataframe=filter_dataframe,
        inverse=inverse,
        return_node_list=return_node_list,
        recompute_distmat=recompute_distmat,
        update_coords=update_coords,
    )


def extract_interface_subgraph(
    g: nx.Graph,
    interface_list: Optional[List[str]] = None,
    chain_list: Optional[List[str]] = None,
    filter_dataframe: bool = True,
    update_coords: bool = True,
    recompute_distmat: bool = False,
    inverse: bool = False,
    return_node_list: bool = False,
) -> Union[nx.Graph, List[str]]:
    """Extracts a subgraph from a graph of a complexed structure (multiple
    chains).

    If there is an edge between two nodes that are part of different chains it
    is included in the selection. NB - if you want to be precise about the
    interfacial region, you should compute this on the basis of solvent
    accessibility and make the selection with
    :method:`extract_subgraph_from_node_list`.

    :param g: The graph to extract the subgraph from.
    :type g: nx.Graph
    :param interface_list: A list of interface names to extract e.g.
        ``["AB", "CD"]``. Default is ``None``.
    :type interface_list: Optional[List[str]]
    :param chain_list: A list of chain names to extract e.g.
        ``["A", "B"]``. Default is ``None``.
    :type chain_list: Optional[List[str]]
    :param filter_dataframe: Whether to filter the ``pdb_df`` dataframe of the
        graph. Defaults to ``True``.
    :type filter_dataframe: bool
    :param update_coords: Whether to update the coordinates of the graph.
        Defaults to ``True``.
    :type update_coords: bool
    :param recompute_distmat: Whether to recompute the distance matrix of the
        graph. Defaults to ``False``.
    :type recompute_distmat: bool
    :param inverse: Whether to inverse the selection. Defaults to ``False``.
    :type inverse: bool
    :return: The subgraph or node list if ``return_node_list=True``.
    :rtype: Union[nx.Graph, List[str]]
    """
    node_list: List[str] = []
    # Iterate over edges
    for u, v in g.edges():
        u_chain = g.nodes[u]["chain_id"]
        v_chain = g.nodes[v]["chain_id"]

        # If a list of chain selections are given, check if the edge is between
        # chains in the list. Save nodes if the chains are not the same
        if (
            chain_list is not None
            and u_chain in chain_list
            and v_chain in chain_list
            and u_chain != v_chain
        ):
            node_list.extend((u, v))

        # If a list of valid interfaces is provided, check the nodes are in the
        # valid interfaces
        if interface_list is not None:
            case_1 = u_chain + v_chain
            case_2 = v_chain + u_chain
            if case_1 in interface_list or case_2 in interface_list:
                node_list.extend((u, v))

        # In the absence of selection params
        # Save nodes if the chains are not the same
        if (
            chain_list is None
            and interface_list is None
            and u_chain != v_chain
        ):
            node_list.extend((u, v))

    node_list = list(set(node_list))

    log.debug(f"Found {len(node_list)} nodes in the interface subgraph.")

    return extract_subgraph_from_node_list(
        g,
        node_list,
        filter_dataframe=filter_dataframe,
        inverse=inverse,
        return_node_list=return_node_list,
        recompute_distmat=recompute_distmat,
        update_coords=update_coords,
    )


def extract_subgraph(
    g: nx.Graph,
    node_list: Optional[List[str]] = None,
    sequence_positions: Optional[List[int]] = None,
    chains: Optional[List[str]] = None,
    residue_types: Optional[List[str]] = None,
    atom_types: Optional[List[str]] = None,
    bond_types: Optional[List[str]] = None,
    centre_point: Optional[
        Union[np.ndarray, Tuple[float, float, float]]
    ] = None,
    radius: Optional[float] = None,
    ss_elements: Optional[List[str]] = None,
    rsa_threshold: Optional[float] = None,
    k_hop_central_node: Optional[str] = None,
    k_hops: Optional[int] = None,
    k_only: Optional[bool] = None,
    filter_dataframe: bool = True,
    update_coords: bool = True,
    recompute_distmat: bool = False,
    inverse: bool = False,
    return_node_list: bool = False,
) -> Union[nx.Graph, List[str]]:
    """Extracts a subgraph from a graph based on a list of nodes, sequence
    positions, chains, residue types, atom types, centre point and radius.

    :param g: The graph to extract the subgraph from.
    :type g: nx.Graph
    :param node_list: List of nodes to extract specified by their ``node_id``.
        Defaults to ``None``.
    :type node_list: List[str], optional
    :param sequence_positions: The sequence positions to extract. Defaults to
        ``None``.
    :type sequence_positions: List[int], optional
    :param chains: The chain(s) to extract. Defaults to ``None``.
    :type chains: List[str], optional
    :param residue_types: List of allowable residue types (3 letter residue
        names). Defaults to ``None``.
    :type residue_types: List[str], optional
    :param atom_types: List of allowable atom types. Defaults to ``None``.
    :type atom_types: List[str], optional
    :param centre_point: The centre point to extract the subgraph from. Defaults
        to ``None``.
    :type centre_point: Union[np.ndarray, Tuple[float, float, float]], optional
    :param radius: The radius to extract the subgraph from.
        Defaults to ``None``.
    :type radius: float, optional
    :param ss_elements: List of secondary structure elements to extract.
        ``["H", "B", "E", "G", "I", "T", "S", "-"]`` corresponding to Alpha
        helix, Beta bridge, Strand, Helix-3, Helix-5, Turn, Bend, None.
        Defaults to ``None``.
    :type ss_elements: List[str], optional
    :param rsa_threshold: The threshold to use for the RSA. Defaults to
        ``None``.
    :type rsa_threshold: float, optional
    :param central_node: The central node to extract the subgraph from.
        Defaults to ``None``.
    :type central_node: str, optional
    :param k: The number of hops to extract.
    :type k: int
    :param k_only: Whether to only extract the exact k-hop subgraph (e.g.
        include 2-hop neighbours in 5-hop graph). Defaults to ``False``.
    :type k_only: bool
    :param filter_dataframe: Whether to filter the ``pdb_df`` dataframe of the
        graph. Defaults to ``True``. Defaults to ``None``.
    :type filter_dataframe: bool, optional
    :param update_coords: Whether to update the coordinates of the graph.
        Defaults to ``True``.
    :type update_coords: bool
    :param recompute_distmat: Whether to recompute the distance matrix of the
        graph. Defaults to ``False``.
    :type recompute_distmat: bool
    :param inverse: Whether to inverse the selection. Defaults to ``False``.
    :type inverse: bool, optional
    :return: The subgraph or node list if ``return_node_list=True``.
    :rtype: Union[nx.Graph, List[str]]
    """
    if node_list is None:
        node_list = []

    if sequence_positions is not None:
        node_list += extract_subgraph_by_sequence_position(
            g, sequence_positions, return_node_list=True
        )

    if chains is not None:
        node_list += extract_subgraph_from_chains(
            g, chains, return_node_list=True
        )

    if residue_types is not None:
        node_list += extract_subgraph_from_residue_types(
            g, residue_types, return_node_list=True
        )

    if atom_types is not None:
        node_list += extract_subgraph_from_atom_types(
            g, atom_types, return_node_list=True
        )

    if bond_types is not None:
        node_list += extract_subgraph_by_bond_type(
            g, bond_types, return_node_list=True
        )

    if centre_point is not None and radius is not None:
        node_list += extract_subgraph_from_point(
            g, centre_point, radius, return_node_list=True
        )

    if ss_elements is not None:
        node_list += extract_subgraph_from_secondary_structure(
            g, ss_elements, return_node_list=True
        )

    if rsa_threshold is not None:
        node_list += extract_surface_subgraph(
            g, rsa_threshold, return_node_list=True
        )

    if k_hop_central_node is not None and k_hops and k_only is not None:
        node_list += extract_k_hop_subgraph(
            g, k_hop_central_node, k_hops, k_only, return_node_list=True
        )

    node_list = list(set(node_list))

    return extract_subgraph_from_node_list(
        g,
        node_list,
        filter_dataframe=filter_dataframe,
        inverse=inverse,
        return_node_list=return_node_list,
        recompute_distmat=recompute_distmat,
        update_coords=update_coords,
    )



"""Utilities for manipulating tensor representations of proteins."""

# Graphein
# Author: Arian Jamasb <arian@jamasb.io>
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein


from .data import Protein
from .io import protein_df_to_tensor



"""Tests for graphein.protein.tensor.reconstruction."""

# Graphein
# Author: Arian Jamasb <arian@jamasb.io>
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein

import pytest

from graphein.protein.tensor.geometry import kabsch
from graphein.protein.tensor.reconstruction import dist_mat_to_coords

try:
    import torch

    TORCH_AVAIL = True
except ImportError:
    TORCH_AVAIL = False


@pytest.mark.skipif(not TORCH_AVAIL, reason="PyTorch not available")
def test_dist_mat_to_coords():
    # Test that the distance matrix is recovered from the coordinates, with a
    # small error.
    for _ in range(10):
        coords = torch.rand((10, 3))
        d = torch.cdist(coords, coords)
        X = dist_mat_to_coords(d)
        assert torch.allclose(d, torch.cdist(X, X), atol=1e-4)
        X_aligned = kabsch(X, coords)
        assert torch.allclose(coords, X_aligned, atol=1e-4)



import networkx as nx
import numpy as np

from graphein.utils.dependencies import import_message

try:
    import prody
except ImportError:
    message = import_message(
        "graphein.protein.features.graph.dynamics", "prody", "prody", True
    )


def add_modes(
    g: nx.Graph, n_modes: int = 6, add_separate: bool = False
) -> nx.Graph:
    """
    Computes normal modes for a protein graph.

    Normal modes are accessible with: ``g.graph["nma_eigenvectors"]``.
    Normal mode eigenvalues are accessible with: ``g.graph["nma_eigenvalues"]``.

    :param g: Protein structure graph
    :type g: nx.Graph
    :param n_modes: Number of modes, defaults to 6
    :type n_modes: int
    :return: Graph with modes added as a graph feature
    :rtype: nx.Graph
    """
    coords = g.graph["coords"]
    anm = prody.ANM(g.graph["name"])
    anm.buildHessian(coords)
    anm.calcModes()

    # eig_vecs = torch.Tensor(anm[:n_modes].getEigvecs()).reshape(-1, n_modes, 3)
    eig_vals = anm[:n_modes].getEigvals()

    modes = np.array([anm[m].getArrayNx3() for m in range(n_modes)])
    modes = np.einsum("ijk->jki", modes)

    if add_separate:
        for i, (_, d) in enumerate(g.nodes(data=True)):
            for mode in range(n_modes):
                d[f"nma_eigenvectors_{mode}"] = modes[i, :, mode]
                d[f"nma_eigenvalues_{mode}"] = modes[i, :, mode]

    for i, (_, d) in enumerate(g.nodes(data=True)):
        d["nma_eigenvectors"] = modes[i, :, :]
        d["nma_eigenvalues"] = eig_vals

    return g



"""Utilities for converting Graphein Networks to Geometric Deep Learning formats."""

# %%
# Graphein
# Author: Kexin Huang, Arian Jamasb <arian@jamasb.io>
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein
from __future__ import annotations

from typing import List, Optional

import networkx as nx
import numpy as np
import torch
from loguru import logger as log
from torch_geometric.utils.undirected import to_undirected

from graphein.utils.dependencies import import_message

try:
    import torch
except ImportError:
    import_message(
        submodule="graphein.ml.conversion",
        package="torch",
        pip_install=True,
        conda_channel="pytorch",
    )

try:
    import torch_geometric
    from torch_geometric.data import Data
except ImportError:
    import_message(
        submodule="graphein.ml.conversion",
        package="torch_geometric",
        pip_install=True,
        conda_channel="rusty1s",
    )

try:
    import dgl
except ImportError:
    import_message(
        submodule="graphein.ml.conversion",
        package="dgl",
        pip_install=True,
        conda_channel="dglteam",
    )

try:
    import jax.numpy as jnp
except ImportError:
    import_message(
        submodule="graphein.ml.conversion",
        package="jax",
        pip_install=True,
        conda_channel="conda-forge",
    )
try:
    import jraph
except ImportError:
    import_message(
        submodule="graphein.ml.conversion",
        package="jraph",
        pip_install=True,
        conda_channel="conda-forge",
    )


SUPPORTED_FORMATS = ["nx", "pyg", "dgl", "jraph"]
"""Supported conversion formats.

``"nx"``: NetworkX graph

``"pyg"``: PyTorch Geometric Data object

``"dgl"``: DGL graph

``"Jraph"``: Jraph GraphsTuple
"""

SUPPORTED_VERBOSITY = ["gnn", "default", "all_info"]
"""Supported verbosity levels for preserving graph features in conversion."""


class GraphFormatConvertor:
    """
    Provides conversion utilities between NetworkX Graphs and geometric deep learning library destination formats.
    Currently, we provide support for converstion from ``nx.Graph`` to ``dgl.DGLGraph`` and ``pytorch_geometric.Data``. Supported conversion
    formats can be retrieved from :const:`~graphein.ml.conversion.SUPPORTED_FORMATS`.

    :param src_format: The type of graph you'd like to convert from. Supported formats are available in :const:`~graphein.ml.conversion.SUPPORTED_FORMATS`
    :type src_format: Literal["nx", "pyg", "dgl", "jraph"]
    :param dst_format: The type of graph format you'd like to convert to. Supported formats are available in:
        ``graphein.ml.conversion.SUPPORTED_FORMATS``
    :type dst_format:  Literal["nx", "pyg", "dgl", "jraph"]
    :param verbose: Select from ``"gnn"``, ``"default"``, ``"all_info"`` to determine how much information is preserved (features)
        as some are unsupported by various downstream frameworks
    :type verbose: graphein.ml.conversion.SUPPORTED_VERBOSITY
    :param columns: List of columns in the node features to retain
    :type columns: List[str], optional
    """

    def __init__(
        self,
        src_format: str,
        dst_format: str,
        verbose: SUPPORTED_VERBOSITY = "gnn",
        columns: Optional[List[str]] = None,
    ):
        if (src_format not in SUPPORTED_FORMATS) or (
            dst_format not in SUPPORTED_FORMATS
        ):
            raise ValueError(
                "Please specify from supported format, "
                + "/".join(SUPPORTED_FORMATS)
            )
        self.src_format = src_format
        self.dst_format = dst_format

        # supported_verbose_format = ["gnn", "default", "all_info"]
        if (columns is None) and (verbose not in SUPPORTED_VERBOSITY):
            raise ValueError(
                "Please specify the supported verbose mode ("
                + "/".join(SUPPORTED_VERBOSITY)
                + ") or specify column names!"
            )

        if columns is None:
            if verbose == "gnn":
                columns = [
                    "edge_index",
                    "coords",
                    "name",
                    "node_id",
                ]
            elif verbose == "default":
                columns = [
                    "b_factor",
                    "chain_id",
                    "coords",
                    "edge_index",
                    "kind",
                    "name",
                    "node_id",
                    "residue_name",
                ]
            elif verbose == "all_info":
                columns = [
                    "atom_type",
                    "b_factor",
                    "chain_id",
                    "chain_ids",
                    "config",
                    "coords",
                    "dist_mat",
                    "edge_index",
                    "element_symbol",
                    "kind",
                    "name",
                    "node_id",
                    "node_type",
                    "pdb_df",
                    "raw_pdb_df",
                    "residue_name",
                    "residue_number",
                    "rgroup_df",
                    "sequence_A",
                    "sequence_B",
                ]
        self.columns = columns

        self.type2form = {
            "atom_type": "str",
            "b_factor": "float",
            "chain_id": "str",
            "coords": "np.array",
            "dist_mat": "np.array",
            "element_symbol": "str",
            "node_id": "str",
            "residue_name": "str",
            "residue_number": "int",
            "edge_index": "torch.tensor",
            "kind": "str",
        }

    def convert_nx_to_dgl(self, G: nx.Graph) -> dgl.DGLGraph:
        """
        Converts ``NetworkX`` graph to ``DGL``

        :param G: ``nx.Graph`` to convert to ``DGLGraph``
        :type G: nx.Graph
        :return: ``DGLGraph`` object version of input ``NetworkX`` graph
        :rtype: dgl.DGLGraph
        """
        g = dgl.DGLGraph()
        node_id = list(G.nodes())
        G = nx.convert_node_labels_to_integers(G)

        ## add node level feat

        node_dict = {}
        for i, (_, feat_dict) in enumerate(G.nodes(data=True)):
            for key, value in feat_dict.items():
                if str(key) in self.columns:
                    node_dict[str(key)] = (
                        [value] if i == 0 else node_dict[str(key)] + [value]
                    )

        string_dict = {}
        node_dict_transformed = {}
        for i, j in node_dict.items():
            if i == "coords":
                node_dict_transformed[i] = torch.Tensor(np.asarray(j)).type(
                    "torch.FloatTensor"
                )
            elif i == "dist_mat":
                node_dict_transformed[i] = torch.Tensor(
                    np.asarray(j[0].values)
                ).type("torch.FloatTensor")
            elif self.type2form[i] == "str":
                string_dict[i] = j
            elif self.type2form[i] in ["float", "int"]:
                node_dict_transformed[i] = torch.Tensor(np.array(j))
        g.add_nodes(
            len(node_id),
            node_dict_transformed,
        )

        edge_dict = {}
        edge_index = torch.LongTensor(list(G.edges)).t().contiguous()

        # add edge level features
        for i, (_, _, feat_dict) in enumerate(G.edges(data=True)):
            for key, value in feat_dict.items():
                if str(key) in self.columns:
                    edge_dict[str(key)] = (
                        list(value)
                        if i == 0
                        else edge_dict[str(key)] + list(value)
                    )

        edge_transform_dict = {}
        for i, j in node_dict.items():
            if self.type2form[i] == "str":
                string_dict[i] = j
            elif self.type2form[i] in ["float", "int"]:
                edge_transform_dict[i] = torch.Tensor(np.array(j))
        g.add_edges(edge_index[0], edge_index[1], edge_transform_dict)

        # add graph level features
        graph_dict = {
            str(feat_name): [G.graph[feat_name]]
            for feat_name in G.graph
            if str(feat_name) in self.columns
        }

        return g

    def convert_nx_to_pyg(self, G: nx.Graph) -> Data:
        """
        Converts ``NetworkX`` graph to ``pytorch_geometric.data.Data`` object. Requires ``PyTorch Geometric`` (https://pytorch-geometric.readthedocs.io/en/latest/) to be installed.

        :param G: ``nx.Graph`` to convert to PyTorch Geometric ``Data`` object
        :type G: nx.Graph
        :return: ``Data`` object containing networkx graph data
        :rtype: pytorch_geometric.data.Data
        """

        # Initialise dict used to construct Data object & Assign node ids as a feature
        data = {"node_id": list(G.nodes())}
        G = nx.convert_node_labels_to_integers(G)

        # Construct Edge Index
        edge_index = (
            torch.LongTensor(list(G.edges)).t().contiguous().view(2, -1)
        )

        # Add node features
        node_feature_names = G.nodes(data=True)[0].keys()
        for i, (_, feat_dict) in enumerate(G.nodes(data=True)):
            for key, value in feat_dict.items():
                key = str(key)
                if key in self.columns:
                    if i == 0:
                        data[key] = []
                    data[key].append(value)

        # Add edge features
        edge_list = list(G.edges(data=True))
        edge_feature_names = edge_list[0][2].keys() if edge_list else []
        edge_feature_names = list(
            filter(
                lambda x: x in self.columns and x != "kind", edge_feature_names
            )
        )
        for i, (_, _, feat_dict) in enumerate(G.edges(data=True)):
            for key, value in feat_dict.items():
                key = str(key)
                if key in self.columns or key == "kind":
                    if i == 0:
                        data[key] = []
                    data[key].append(value)

        # Add graph-level features
        for feat_name in G.graph:
            if str(feat_name) in self.columns:
                if str(feat_name) not in node_feature_names:
                    data[str(feat_name)] = G.graph[feat_name]
        if "edge_index" in self.columns:
            data["edge_index"] = edge_index

        # Split edge index by edge kind
        kind_strs = np.array(
            list(map(lambda x: "_".join(x), data.get("kind", [])))
        )
        for kind in set(kind_strs):
            key = f"edge_index_{kind}"
            if key in self.columns:
                mask = kind_strs == kind
                data[key] = edge_index[:, mask]
        if "kind" not in self.columns and data.get("kind"):
            del data["kind"]

        # Convert everything possible to torch.Tensors
        for key, val in data.items():
            try:
                if not isinstance(val, torch.Tensor):
                    data[key] = torch.tensor(np.array(val))
            except Exception as e:
                log.warning(e)
                pass

        # Construct PyG data
        data = Data.from_dict(data)
        data.num_nodes = G.number_of_nodes()

        # Symmetrize if undirected
        if not G.is_directed():
            # Edge index and edge features
            edge_index, edge_features = to_undirected(
                edge_index,
                [getattr(data, attr) for attr in edge_feature_names],
                data.num_nodes,
            )
            if "edge_index" in self.columns:
                data.edge_index = edge_index
            for attr, val in zip(edge_feature_names, edge_features):
                setattr(data, attr, val)

            # Edge indices of different kinds
            for kind in set(kind_strs):
                key = f"edge_index_{kind}"
                if key in self.columns:
                    edge_index_kind = to_undirected(
                        getattr(data, key), num_nodes=data.num_nodes
                    )
                    setattr(data, key, edge_index_kind)

        return data

    @staticmethod
    def convert_nx_to_nx(G: nx.Graph) -> nx.Graph:
        """
        Converts NetworkX graph (``nx.Graph``) to NetworkX graph (``nx.Graph``) object. Redundant - returns itself.

        :param G: NetworkX Graph
        :type G: nx.Graph
        :return: NetworkX Graph
        :rtype: nx.Graph
        """
        return G

    @staticmethod
    def convert_dgl_to_nx(G: dgl.DGLGraph) -> nx.Graph:
        """
        Converts a DGL Graph (``dgl.DGLGraph``) to a NetworkX (``nx.Graph``) object. Preserves node and edge attributes.

        :param G: ``dgl.DGLGraph`` to convert to ``NetworkX`` graph.
        :type G: dgl.DGLGraph
        :return: NetworkX graph object.
        :rtype: nx.Graph
        """
        node_attrs = G.node_attr_schemes().keys()
        edge_attrs = G.edge_attr_schemes().keys()
        return dgl.to_networkx(G, node_attrs, edge_attrs)

    @staticmethod
    def convert_pyg_to_nx(G: Data) -> nx.Graph:
        """Converts PyTorch Geometric ``Data`` object to NetworkX graph (``nx.Graph``).

        :param G: Pytorch Geometric Data.
        :type G: torch_geometric.data.Data
        :returns: NetworkX graph.
        :rtype: nx.Graph
        """
        return torch_geometric.utils.to_networkx(G)

    def convert_nx_to_jraph(self, G: nx.Graph) -> jraph.GraphsTuple:
        """Converts NetworkX graph (``nx.Graph``) to Jraph GraphsTuple graph. Requires ``jax`` and ``Jraph``.

        :param G: Networkx graph to convert.
        :type G: nx.Graph
        :return: Jraph GraphsTuple graph.
        :rtype: jraph.GraphsTuple
        """
        G = nx.convert_node_labels_to_integers(G)

        n_node = len(G)
        n_edge = G.number_of_edges()
        edge_list = list(G.edges())
        senders, receivers = zip(*edge_list)
        senders, receivers = jnp.array(senders), jnp.array(receivers)

        # Add node features
        node_features = {}
        for i, (_, feat_dict) in enumerate(G.nodes(data=True)):
            for key, value in feat_dict.items():
                if str(key) in self.columns:
                    # node_features[str(key)] = (
                    #    [value]
                    #    if i == 0
                    #    else node_features[str(key)] + [value]
                    # )
                    feat = (
                        [value]
                        if i == 0
                        else node_features[str(key)] + [value]
                    )
                    try:
                        feat = torch.tensor(feat)
                        node_features[str(key)] = feat
                    except TypeError:
                        node_features[str(key)] = feat

        # Add edge features
        edge_features = {}
        for i, (_, _, feat_dict) in enumerate(G.edges(data=True)):
            for key, value in feat_dict.items():
                if str(key) in self.columns:
                    edge_features[str(key)] = (
                        list(value)
                        if i == 0
                        else edge_features[str(key)] + list(value)
                    )

        # Add graph features
        global_context = {
            str(feat_name): [G.graph[feat_name]]
            for feat_name in G.graph
            if str(feat_name) in self.columns
        }

        return jraph.GraphsTuple(
            nodes=node_features,
            senders=senders,
            receivers=receivers,
            edges=edge_features,
            n_node=n_node,
            n_edge=n_edge,
            globals=global_context,
        )

    def __call__(self, G: nx.Graph):
        nx_g = eval("self.convert_" + self.src_format + "_to_nx(G)")
        dst_g = eval("self.convert_nx_to_" + self.dst_format + "(nx_g)")
        return dst_g


def convert_nx_to_pyg_data(G: nx.Graph) -> Data:
    # Initialise dict used to construct Data object
    data = {"node_id": list(G.nodes())}

    G = nx.convert_node_labels_to_integers(G)

    # Construct Edge Index
    edge_index = torch.LongTensor(list(G.edges)).t().contiguous()

    # Add node features
    for i, (_, feat_dict) in enumerate(G.nodes(data=True)):
        for key, value in feat_dict.items():
            data[str(key)] = [value] if i == 0 else data[str(key)] + [value]

    # Add edge features
    for i, (_, _, feat_dict) in enumerate(G.edges(data=True)):
        for key, value in feat_dict.items():
            data[str(key)] = (
                list(value) if i == 0 else data[str(key)] + list(value)
            )

    # Add graph-level features
    for feat_name in G.graph:
        data[str(feat_name)] = [G.graph[feat_name]]

    data["edge_index"] = edge_index.view(2, -1)
    data = Data.from_dict(data)
    data.num_nodes = G.number_of_nodes()

    return data



"""Utilities for parsing the TRRUST database."""

# %%
# Graphein
# Author: Ramon Vinas, Arian Jamasb <arian@jamasb.io>
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein

import os
from functools import lru_cache
from pathlib import Path
from typing import Callable, List, Optional

import pandas as pd
import wget
from loguru import logger as log

from graphein.utils.utils import filter_dataframe


def _download_TRRUST(root_dir: Optional[Path] = None) -> str:
    """
    Downloads TRRUST from
    https://www.grnpedia.org/trrust/data/trrust_rawdata.human.tsv

    :param root_dir: Path to desired output directory to download TRRUST to.
        Defaults to None (downloads to graphein/datasets/trrust/)
    :type root_dir: pathlib.Path, optional
    :returns: Path to downloaded TRRUST Dataset.
    :rtype: str
    """
    url = "https://www.grnpedia.org/trrust/data/trrust_rawdata.human.tsv"

    if root_dir is None:
        root_dir = Path(__file__).parent.parent.parent
    trrust_dir = f"{root_dir}/datasets/trrust"
    Path(trrust_dir).mkdir(parents=False, exist_ok=True)
    file = f"{trrust_dir}/human.tsv"

    # Download data
    if not os.path.exists(file):
        log.info("Downloading TRRUST ...")
        wget.download(url, file)

    return file


@lru_cache()
def load_TRRUST(root_dir: Optional[Path] = None) -> pd.DataFrame:
    """
    Loads the TRRUST datafile. If file not found, it is downloaded.

    :param root_dir: Root directory path to either find or download TRRUST
    :type root_dir: pathlib.Path, optional
    :returns: TRRUST database as a dataframe
    :rtype: pd.DataFrame
    """
    file = _download_TRRUST(root_dir)

    return pd.read_csv(
        file,
        delimiter="\t",
        header=None,
        names=["g1", "g2", "regtype", "references"],
    )


def parse_TRRUST(
    gene_list: List[str], root_dir: Optional[Path] = None
) -> pd.DataFrame:
    """
    Parser for TRRUST regulatory interactions. If the TRRUST dataset is not
    found in the specified root_dir, it is downloaded

    :param gene_list: List of gene identifiers to restrict dataframe to.
    :type gene_list: List[str]
    :param root_dir: Root directory path to either find or download TRRUST.
        Defaults to ``None`` (downloads dataset to ``graphein/datasets/trrust``)
    :type root_dir: pathlib.Path, optional
    :returns: Pandas DataFrame with the regulatory interactions between genes
        in the gene list.
    :rtype: pd.DataFrame
    """
    df = load_TRRUST(root_dir=root_dir)
    # Select input genes
    df = df[df["g1"].isin(gene_list) & df["g2"].isin(gene_list)]

    return df


def filter_TRRUST(
    df: pd.DataFrame, funcs: Optional[List[Callable]]
) -> pd.DataFrame:
    """
    Filters results of TRRUST call according to user kwargs.

    :param df: Source specific Pandas dataframe (TRRUST) with results of the
        API call
    :type df: pd.DataFrame
    :param funcs: User functions to filter the results.
    :type funcs: List[Callable]
    :return: Source specific Pandas dataframe with filtered results
    :rtype: pd.DataFrame
    """
    if funcs is not None:
        df = filter_dataframe(df, funcs)

    return df


def standardise_TRRUST(df: pd.DataFrame) -> pd.DataFrame:
    """
    Filters results of TRRUST call by providing a list of
    user-defined functions that accept a dataframe and return a dataframe.

    :param df: pd.Dataframe to filter. Must contain columns:
        ``["g1", "g2", "regtype"]``.
    :type df: pd.DataFrame
    :param funcs: List of functions that carry out dataframe processing.
    :type funcs: List[Callable]
    :return: Processed DataFrame.
    :rtype: pd.DataFrame
    """
    # Rename & delete columns
    df = df[["g1", "g2", "regtype"]]

    # Rename type of regulatory interaction
    df["regtype"].replace(
        {"Activation": "+", "Repression": "-", "Unknown": "?"}, inplace=True
    )

    # Add source column
    df["source"] = "TRRUST"

    return df


def TRRUST_df(
    gene_list: List[str], filtering_funcs: Optional[List[Callable]] = None
) -> pd.DataFrame:
    """
    Generates standardised dataframe with TRRUST protein-protein interactions,
    filtered according to user's input.

    :param gene_list: List of gene identifiers to restrict dataframe to.
    :type gene_list: List[str]
    :param filtering_funcs: Functions with which to filter the dataframe.
    :type filtering_funcs: List[Callable]
    :return: Standardised dataframe with TRRUST interactions.
    :rtype: pd.DataFrame
    """
    df = parse_TRRUST(gene_list=gene_list)
    df = filter_TRRUST(df, filtering_funcs)
    df = standardise_TRRUST(df)

    return df


if __name__ == "__main__":
    df = TRRUST_df(["AATF", "MYC", "USF1", "SP1", "TP53", "DUSP1"])
    print(df.head())



"""Protein graph construction module."""

from .config import *
from .edges import *
from .features import *
from .graphs import *
from .resi_atoms import *
from .subgraphs import *
from .utils import *
from .visualisation import (
    plot_distance_landscape,
    plot_distance_matrix,
    plot_protein_structure_graph,
    plotly_protein_structure_graph,
)

try:
    from .visualisation import plot_chord_diagram
except ImportError:
    pass

try:
    from .meshes import *
    from .visualisation import plot_pointcloud
except ImportError:
    pass



from __future__ import print_function

import os
import subprocess
import tempfile
import time

from ..dependencies import requires_external_dependencies
from .compat import Server

HOST = os.environ.get("PYMOL_RPCHOST", "localhost")
PORT = 9123


class MolViewer(object):
    def __init__(self, host=HOST, port=PORT):
        self.host = host
        self.port = int(port)
        self._process = None

    def __del__(self):
        self.stop()

    def __getattr__(self, key):
        if not self._process_is_running():
            self.start(["-cKQ"])

        return getattr(self._server, key)

    def _process_is_running(self):
        return self._process is not None and self._process.poll() is None

    @requires_external_dependencies("pymol")
    def start(self, args=("-Q",), exe="pymol"):
        """Start the PyMOL RPC server and connect to it
        Start simple GUI (-xi), suppress all output (-Q):
            >>> viewer.start(["-xiQ"])
        Start headless (-cK), with some output (-q):
            >>> viewer.start(["-cKq"])
        """
        if self._process_is_running():
            print("A PyMOL RPC server is already running.")
            return

        assert isinstance(args, (list, tuple))

        self._process = subprocess.Popen([exe, "-R"] + list(args))

        self._server = Server(uri="http://%s:%d/RPC2" % (self.host, self.port))

        # wait for the server
        while True:
            try:
                self._server.bg_color("white")
                break
            except IOError:
                time.sleep(0.1)

    def stop(self):
        if self._process_is_running():
            self._process.terminate()

    def display(self, width=0, height=0, ray=False, timeout=120):
        """Display PyMol session
        :param width: width in pixels (0 uses current viewport)
        :param height: height in pixels (0 uses current viewport)
        :param ray: use ray tracing (if running PyMOL headless, this parameter
        has no effect and ray tracing is always used)
        :param timeout: timeout in seconds
        Returns
        -------
        fig : IPython.display.Image
        """
        from IPython.display import Image, display
        from ipywidgets import IntProgress

        progress_max = int((timeout * 20) ** 0.5)
        progress = None
        filename = tempfile.mktemp(".png")

        try:
            self._server.png(filename, width, height, -1, int(ray))

            for i in range(1, progress_max):
                if os.path.exists(filename):
                    break

                if progress is None:
                    progress = IntProgress(min=0, max=progress_max)
                    display(progress)

                progress.value += 1
                time.sleep(i / 10.0)

            if not os.path.exists(filename):
                raise RuntimeError("timeout exceeded")

            return Image(filename)
        finally:
            if progress is not None:
                progress.close()

            try:
                os.unlink(filename)
            except:
                pass


# Create a default instance for convenience
viewer = MolViewer()



"""Utilities for working with graph objects."""

# Graphein
# Author: Arian Jamasb <arian@jamasb.io>, Eric Ma
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein
from __future__ import annotations

import functools
import inspect
import platform
import subprocess
import sys
import warnings
from typing import Any, Callable, Iterable, List

import networkx as nx
import numpy as np
import pandas as pd
import xarray as xr
from Bio.Data.IUPACData import protein_letters_3to1

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


def onek_encoding_unk(
    x: Iterable[Any], allowable_set: List[Any]
) -> List[bool]:
    """
    Function for perfroming one hot encoding

    :param x: values to one-hot
    :type x: Iterable[Any]
    :param allowable_set: set of options to encode
    :type allowable_set: List[Any]
    :return: one-hot encoding as list
    :rtype: List[bool]
    """
    # if x not in allowable_set:
    #    x = allowable_set[-1]
    return [x == s for s in allowable_set]


def filter_dataframe(df: pd.DataFrame, funcs: List[Callable]) -> pd.DataFrame:
    """
    Applies transformation functions to a dataframe. Each function in ``funcs`` must accept a ``pd.DataFrame`` and return a ``pd.DataFrame``.

    Additional parameters can be provided by using partial functions.

    :param df: Dataframe to apply transformations to.
    :type df: pd.DataFrame
    :param funcs: List of transformation functions.
    :type funcs: List[Callable]
    :rtype: nx.Graph
    """
    for func in funcs:
        func(df)
    return df


def annotate_graph_metadata(G: nx.Graph, funcs: List[Callable]) -> nx.Graph:
    """
    Annotates graph with graph-level metadata

    :param G: Graph on which to add graph-level metadata to
    :type G: nx.Graph
    :param funcs: List of graph metadata annotation functions
    :type funcs: List[Callable]
    :return: Graph on which with node metadata added
    :rtype: nx.Graph
    """
    for func in funcs:
        func(G)
    return G


def annotate_edge_metadata(G: nx.Graph, funcs: List[Callable]) -> nx.Graph:
    """
    Annotates Graph edges with edge metadata. Each function in ``funcs`` must take the three arguments ``u``, ``v`` and ``d``, where ``u`` and ``v`` are the nodes of the edge, and ``d`` is the edge data dictionary.

    Additional parameters can be provided by using partial functions.

    :param G: Graph to add edge metadata to
    :type G: nx.Graph
    :param funcs: List of edge metadata annotation functions
    :type funcs: List[Callable]
    :return: Graph with edge metadata added
    :rtype: nx.Graph
    """
    for func in funcs:
        for u, v, d in G.edges(data=True):
            func(u, v, d)
    return G


def annotate_node_metadata(G: nx.Graph, funcs: List[Callable]) -> nx.Graph:
    """
    Annotates nodes with metadata. Each function in ``funcs`` must take two arguments ``n`` and ``d``, where ``n`` is the node and ``d`` is the node data dictionary. Some functions, like esm residue embeddings, should be used as graph metadata functions as they require access to the whole sequence.

    Additional parameters can be provided by using partial functions.

    :param G: Graph to add node metadata to
    :type G: nx.Graph
    :param funcs: List of node metadata annotation functions
    :type funcs: List[Callable]
    :return: Graph with node metadata added
    :rtype: nx.Graph
    """

    for func in funcs:
        for n, d in G.nodes(data=True):
            try:
                func(n, d)
            except (TypeError, AttributeError) as e:
                raise type(e)(
                    str(e)
                    + "Please ensure that provided node metadata functions match the f(n: str, d: Dict) function signature, where n is the node ID and d is the node data dictionary "
                ).with_traceback(sys.exc_info()[2])
    return G


def annotate_node_features(G: nx.Graph, funcs: List[Callable]) -> nx.Graph:
    """
    Annotates nodes with features data. Note: passes whole graph to function.

    :param G: Graph to add node features to
    :type G: nx.Graph
    :param funcs: List of node feature annotation functions
    :type funcs: List[Callable]
    :return: Graph with node features added
    :rtype: nx.Graph
    """
    for func in funcs:
        func(G)
    return G


def compute_edges(G: nx.Graph, funcs: List[Callable]) -> nx.Graph:
    """
    Computes edges for an Graph from a list of edge construction functions. Each func in ``funcs`` must take an ``nx.Graph`` and return an ``nx.Graph``.

    :param G: Graph to add features to
    :type G: nx.Graph
    :param funcs: List of edge construction functions
    :type funcs: List[Callable]
    :return: Graph with edges added
    :rtype: nx.Graph
    """
    for func in funcs:
        func(G)
    return G


def generate_feature_dataframe(
    G: nx.Graph,
    funcs: List[Callable],
    return_array=False,
) -> pd.DataFrame:
    """
    Return a pandas DataFrame representation of node metadata.

    ``funcs`` has to be list of callables whose signature is

        f(n, d) -> pd.Series

    where ``n`` is the graph node,
    ``d`` is the node metadata dictionary.
    The function must return a pandas Series whose name is the node.

    Example function:

    .. code-block:: python

        def x_vec(n: Hashable, d: Dict[Hashable, Any]) -> pd.Series:
            return pd.Series({"x_coord": d["x_coord"]}, name=n)

    One fairly strong assumption is that each func
    has all the information it needs to act
    stored on the metadata dictionary.
    If you need to reference an external piece of information,
    such as a dictionary to look up values,
    set up the function to accept the dictionary,
    and use ``functools.partial``
    to "reduce" the function signature to just ``(n, d)``.
    An example below:

    .. code-block:: python

        from functools import partial
        def get_molweight(n, d, mw_dict):
            return pd.Series({"mw": mw_dict[d["amino_acid"]]}, name=n)

        mw_dict = {"PHE": 165, "GLY": 75, ...}
        get_molweight_func = partial(get_molweight, mw_dict=mw_dict)

        generate_feature_dataframe(G, [get_molweight_func])

    The ``name=n`` piece is important;
    the ``name`` becomes the row index in the resulting dataframe.

    The series that is returned from each function
    need not only contain one key-value pair.
    You can have two or more, and that's completely fine;
    each key becomes a column in the resulting dataframe.

    A key design choice: We default to returning DataFrames,
    to make inspecting the data easy,
    but for consumption in tensor libraries,
    you can turn on returning a NumPy array
    by switching ``return_array=True``.


    :param G: A NetworkX-compatible graph object.
    :type G: nx.Graph
    :param funcs: A list of functions.
    :type funcs: List[Callable]
    :param return_array: Whether or not to return
        a NumPy array version of the data.
        Useful for consumption in tensor libs, like PyTorch or JAX.
    :type return_array: bool
    :return: pandas DataFrame representation of node metadata.
    :rtype: pd.DataFrame
    """
    matrix = []
    for n, d in G.nodes(data=True):
        series = []
        for func in funcs:
            res = func(n, d)
            if res.name != n:
                raise NameError(
                    f"function {func.__name__} returns a series "
                    "that is not named after the node."
                )
            series.append(res)
        matrix.append(pd.concat(series))

    df = pd.DataFrame(matrix)
    if return_array:
        return df.values
    return df


def format_adjacency(G: nx.Graph, adj: np.ndarray, name: str) -> xr.DataArray:
    """
    Format adjacency matrix nicely.

    Intended to be used when computing an adjacency-like matrix
    of a graph object ``G``.
    For example, in defining a func:

    .. code-block:: python

        def my_adj_matrix_func(G):
            adj = some_adj_func(G)
            return format_adjacency(G, adj, "xarray_coord_name")

    **Assumptions**

    #. ``adj`` should be a 2D matrix of shape ``(n_nodes, n_nodes)``
    #. ``name`` is something that is unique amongst all names used
    in the final adjacency tensor.

    :param G: NetworkX-compatible Graph
    :type param: nx.Graph
    :param adj: 2D numpy array of shape ``(n_nodes, n_nodes)``
    :type adj: np.ndarray
    :param name: A unique name for the kind of adjacency matrix
        being constructed.
        Gets used in xarray as a coordinate in the ``"name"`` dimension.
    :type name: str
    :returns: An XArray DataArray of shape ``(n_nodes, n_nodes, 1)``
    :rtype: xr.DataArray
    """
    expected_shape = (len(G), len(G))
    if adj.shape != expected_shape:
        raise ValueError(
            "Adjacency matrix is not shaped correctly, "
            f"should be of shape {expected_shape}, "
            f"instead got shape {adj.shape}."
        )
    adj = np.expand_dims(adj, axis=-1)
    nodes = list(G.nodes())
    return xr.DataArray(
        adj,
        dims=["n1", "n2", "name"],
        coords={"n1": nodes, "n2": nodes, "name": [name]},
    )


def generate_adjacency_tensor(
    G: nx.Graph, funcs: List[Callable], return_array=False
) -> xr.DataArray:
    """
    Generate adjacency tensor for a graph.

    Uses the collection of functions in ``funcs``
    to build an xarray DataArray
    that houses the resulting "adjacency tensor".

    A key design choice:
    We default to returning xarray DataArrays,
    to make inspecting the data easy,
    but for consumption in tensor libraries,
    you can turn on returning a NumPy array
    by switching ``return_array=True``.

    :param G: NetworkX Graph.
    :type G: nx.Graph
    :param funcs: A list of functions that take in G
        and return an xr.DataArray
    :type funcs: List[Callable]
    :returns: xr.DataArray,
        which is of shape ``(n_nodes, n_nodes, n_funcs)``.
    :rtype: xr.DataArray
    """
    mats = [func(G) for func in funcs]
    da = xr.concat(mats, dim="name")
    if return_array:
        return da.data
    return da


def protein_letters_3to1_all_caps(amino_acid: str) -> str:
    """
    Converts capitalised 3 letter amino acid code to single letter. Not provided in default biopython.

    :param amino_acid: Capitalised 3-letter amino acid code (eg. ``"GLY"``)
    :type amino_acid: str
    :returns: Single-letter amino acid code
    :rtype: str
    """
    amino_acid = amino_acid[0] + amino_acid[1:].lower()
    return protein_letters_3to1[amino_acid]


def ping(host: str) -> bool:
    """
    Returns ``True`` if host (str) responds to a ping request.
    Remember that a host may not respond to a ping (ICMP) request even if the host name is valid.

    :param host: IP or hostname
    :type host: str
    :returns: ``True`` if host responds to a ping request.
    :rtype: bool
    """

    # Option for the number of packets as a function of
    param = "-n" if platform.system().lower() == "windows" else "-c"

    # Building the command. Ex: "ping -c 1 google.com"
    command = ["ping", param, "1", host]

    return subprocess.call(command) == 0


def parse_aggregation_type(
    aggregation_type: Literal["sum", "mean", "max", "min", "median"],
) -> Callable:
    """Returns an aggregation function by name

    :param aggregation_type: One of: ``["max", "min", "mean", "median", "sum"]``.
    :type aggregration_type: Literal["sum", "mean", "max", "min", "median"]
    :returns: NumPy aggregation function.
    :rtype: Callable
    :raises ValueError: if aggregation type is not supported.
    """
    if aggregation_type == "max":
        func = np.max
    elif aggregation_type == "min":
        func = np.min
    elif aggregation_type == "mean":
        func = np.mean
    elif aggregation_type == "median":
        func = np.median
    elif aggregation_type == "sum":
        func = np.sum
    else:
        raise ValueError(
            f"Unsupported aggregator: {aggregation_type}."
            f" Please use min, max, mean, median, sum"
        )
    return func


_string_types = (type(b""), type(""))


def deprecated(reason):
    """
    This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used.
    """

    if isinstance(reason, _string_types):

        def decorator(func1):
            if inspect.isclass(func1):
                fmt1 = "Call to deprecated class {name} ({reason})."
            else:
                fmt1 = "Call to deprecated function {name} ({reason})."

            @functools.wraps(func1)
            def new_func1(*args, **kwargs):
                warnings.simplefilter("always", DeprecationWarning)
                warnings.warn(
                    fmt1.format(name=func1.__name__, reason=reason),
                    category=DeprecationWarning,
                    stacklevel=2,
                )
                warnings.simplefilter("default", DeprecationWarning)
                return func1(*args, **kwargs)

            return new_func1

        return decorator

    elif inspect.isclass(reason) or inspect.isfunction(reason):
        func2 = reason

        if inspect.isclass(func2):
            fmt2 = "Call to deprecated class {name}."
        else:
            fmt2 = "Call to deprecated function {name}."

        @functools.wraps(func2)
        def new_func2(*args, **kwargs):
            warnings.simplefilter("always", DeprecationWarning)
            warnings.warn(
                fmt2.format(name=func2.__name__),
                category=DeprecationWarning,
                stacklevel=2,
            )
            warnings.simplefilter("default", DeprecationWarning)
            return func2(*args, **kwargs)

        return new_func2

    else:
        raise TypeError(repr(type(reason)))



"""Visualisation utils for ML."""

from __future__ import annotations

import itertools
from typing import Iterable, Optional, Tuple

import matplotlib.pyplot as plt
import networkx as nx
import plotly.graph_objects as go

from graphein.protein.features.nodes.geometry import VECTOR_FEATURE_NAMES
from graphein.protein.visualisation import add_vector_to_plot
from graphein.utils.dependencies import import_message

from ..protein.visualisation import plotly_protein_structure_graph
from .conversion import GraphFormatConvertor

try:
    from torch_geometric.data import Data
except ImportError:
    import_message(
        submodule="graphein.ml.conversion",
        package="torch_geometric",
        pip_install=True,
        conda_channel="pyg",
    )

try:
    import torch
except ImportError:
    import_message(
        submodule="graphein.ml.visualisation",
        package="torch",
        pip_install=True,
        conda_channel="pytorch",
    )


def plot_pyg_data(
    x: Data,
    node_colour_tensor: Optional[torch.Tensor] = None,
    edge_colour_tensor: Optional[torch.Tensor] = None,
    plot_title: Optional[str] = None,
    figsize: Tuple[int, int] = (620, 650),
    node_alpha: float = 0.7,
    node_size_min: float = 20.0,
    node_size_multiplier: float = 20.0,
    node_size_feature: str = "degree",
    label_node_ids: bool = True,
    node_colour_map=plt.cm.plasma,
    edge_colour_map=plt.cm.plasma,
    colour_nodes_by: str = "residue_name",
    colour_edges_by: Optional[str] = None,
    node_vector_features: Iterable[str] = tuple(VECTOR_FEATURE_NAMES),
    node_vector_feature_colours: Iterable[str] = ("red", "green", "blue"),
) -> go.Figure:
    """
    Plots protein structure graph from ``torch_geometric.data.Data``
    using plotly.

    This function can be used for logging proteins to e.g. wandb.

    :param x: ``torch_geometric.data.Data`` Protein Structure graph to plot.
    :type x: torch_geometric.data.Data
    :param node_colour_tensor: Tensor of node colours (must match length of
        nodes in graph). If ``None``, ``colour_nodes_by`` will be used. Default
        is ``None``.
    :type node_colour_tensor: torch.Tensor, optional
    :parm edge_colour_tensor: Tensor of edge colours (must match length of edges
        in graph). If ``None``, ``colour_edges_by`` will be used. Default is
        ``None``.
    :type edge_colour_tensor: torch.Tensor, optional
    :param plot_title: Title of plot, defaults to ``None``.
    :type plot_title: str, optional
    :param figsize: Size of figure, defaults to ``(620, 650)``.
    :type figsize: Tuple[int, int]
    :param node_alpha: Controls node transparency, defaults to ``0.7``.
    :type node_alpha: float
    :param node_size_min: Specifies node minimum size. Defaults to ``20.0``.
    :type node_size_min: float
    :param node_size_multiplier: Scales node size by a constant. Node sizes
        reflect degree. Defaults to ``20.0``.
    :type node_size_multiplier: float
    :param node_size_feature: Which feature to scale the node size by. Defaults
        to ``degree``.
    :type node_size_feature: str
    :param label_node_ids: bool indicating whether or not to plot ``node_id``
        labels. Defaults to ``True``.
    :type label_node_ids: bool
    :param node_colour_map: colour map to use for nodes. Defaults to
        ``plt.cm.plasma``.
    :type node_colour_map: plt.cm
    :param edge_colour_map: colour map to use for edges. Defaults to
        ``plt.cm.plasma``.
    :type edge_colour_map: plt.cm
    :param colour_nodes_by: Specifies how to colour nodes. ``"degree"``,
        ``"seq_position"`` or a node feature.
    :type colour_nodes_by: str
    :param colour_edges_by: Specifies how to colour edges. Currently only
        ``"kind"`` or ``None`` are supported.
    :type colour_edges_by: Optional[str]
    :param node_vector_features: Specifies node vector features to visualize. By default all
        present features are plotted.
    :type node_vector_features: Interable[str]
    :param node_vector_feature_colours: Specifies colors of vectors.
    :type node_vector_feature_colours: Interable[str]
    :returns: Plotly Graph Objects plot
    :rtype: go.Figure
    """
    convertor = GraphFormatConvertor(src_format="pyg", dst_format="nx")
    nx_graph = convertor(x)
    node_map = dict(enumerate(x.node_id))
    nx_graph = nx.relabel_nodes(nx_graph, node_map)

    # Add metadata
    nx_graph.name = x.name
    nx_graph.graph["coords"] = x.coords
    nx_graph.graph["dist_mat"] = x.dist_mat

    # Assign coords and seq info to nodes
    for i, (_, d) in enumerate(nx_graph.nodes(data=True)):
        d["chain_id"] = x.node_id[i].split(":")[0]
        d["residue_name"] = x.node_id[i].split(":")[1]
        d["seq_position"] = x.node_id[i].split(":")[2]
        d["coords"] = x.coords[i]
        if node_colour_tensor is not None:
            d["colour"] = float(node_colour_tensor[i])
        for node_vec in node_vector_features:
            if hasattr(x, node_vec):
                d[node_vec] = getattr(x, node_vec)[i]

    # Preprocess edge colours
    if edge_colour_tensor is not None:
        # TODO add edge types
        for i, (_, _, d) in enumerate(nx_graph.edges(data=True)):
            d["colour"] = float(edge_colour_tensor[i])

    # Plot nx graph
    fig = plotly_protein_structure_graph(
        nx_graph,
        plot_title,
        figsize,
        node_alpha,
        node_size_min,
        node_size_multiplier,
        node_size_feature,
        label_node_ids,
        node_colour_map,
        edge_colour_map,
        colour_nodes_by if node_colour_tensor is None else "colour",
        colour_edges_by if edge_colour_tensor is None else "colour",
    )

    # Add vectors to visualize
    node_vector_feature_colours = itertools.cycle(node_vector_feature_colours)
    for node_vec in node_vector_features:
        if hasattr(x, node_vec):
            fig = add_vector_to_plot(
                nx_graph,
                fig,
                node_vec,
                colour=next(node_vector_feature_colours),
                scale=1.5,
            )

    return fig



"""Base Config object for use with RNA Graph Construction."""

# Graphein
# Author: Arian Jamasb <arian@jamasb.io>
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein
from __future__ import annotations

from functools import partial
from pathlib import Path
from typing import Any, Callable, List, Optional, Union

from deepdiff import DeepDiff
from pydantic import BaseModel

from graphein.rna.edges import add_atomic_edges
from graphein.utils.config import PartialMatchOperator, PathMatchOperator


class BpRNAConfig(BaseModel):
    """Config for managing BpRNA
    :param path: Path to bpRNA
    :type path: str
    """

    path: str = "./bpRNA"


class RNAGraphConfig(BaseModel):
    """
    Config Object for RNA Structure Graph Construction.

    :param granularity: Specifies the node types of the graph, defaults to
        ``"rna"`` for atoms as nodes. Other options are ``"rna-centroid"`` or
        any RNA atom name.
    :param verbose: Specifies verbosity of graph creation process.
    :type verbose: bool
    :param rna_df_processing_functions: List of functions that take a
        ``pd.DataFrame`` and return a ``pd.DataFrame``. This allows users to
        define their own series of processing functions for the RNA structure
        DataFrame and override the default sequencing of processing steps
        provided by Graphein. We refer users to our low-level API tutorial for
        more details.
    :type rna_df_processing_functions: Optional[List[Callable]]
    :param edge_construction_functions: List of functions that take an
        ``nx.Graph`` and return an ``nx.Graph`` with desired edges added.
        Prepared edge constructions can be found in :ref:`graphein.rna.edges`
    :type edge_construction_functions: List[Callable]
    :param node_metadata_functions: List of functions that take an ``nx.Graph``
    :type node_metadata_functions: List[Callable], optional
    :param edge_metadata_functions: List of functions that take an
    :type edge_metadata_functions: List[Callable], optional
    :param graph_metadata_functions: List of functions that take an ``nx.Graph``
        and return an ``nx.Graph`` with added graph-level features and metadata.
    :type graph_metadata_functions: List[Callable], optional
    """

    granularity: str = "rna_atom"
    verbose: bool = False
    insertions: bool = False
    keep_hets: List[str] = []

    # Graph construction functions
    edge_construction_functions: List[Union[Callable, str]] = [
        add_atomic_edges
    ]
    rna_df_processing_functions: Optional[List[Callable]] = None
    node_metadata_functions: Optional[List[Union[Callable, str]]] = None
    edge_metadata_functions: Optional[List[Union[Callable, str]]] = None
    graph_metadata_functions: Optional[List[Callable]] = None

    def __eq__(self, other: Any) -> bool:
        """
        Overwrites the BaseModel __eq__ function in order to check more
        specific cases (like partial functions).
        """
        if isinstance(other, RNAGraphConfig):
            return (
                DeepDiff(
                    self,
                    other,
                    custom_operators=[
                        PartialMatchOperator(types=[partial]),
                        PathMatchOperator(types=[Path]),
                    ],
                )
                == {}
            )
        return self.dict() == other
